# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..utils import dict_add, url_merge
from .common import ItemExtractor, ForumExtractor
from .identify import identify_xmb


def dict_url_merge(ref, data, fields):
    for j in fields:
        if len(j) == 1:
            data[j[0]] = url_merge(ref, data[j[0]])
        else:
            data[j[0]][j[1]] = url_merge(ref, data[j[0]][j[1]])


class xmb(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/(.*/)?viewthread\.php\?tid=(\d+)$"),
                    2,
                )
            ]
            self.trim = True

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "xmb-thread", "url": url, "id": int(i_id)}
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                    .title td .nav -style | "%i" / sed "s/.* &raquo; //" decode trim,
                    .path.a [0] td .nav -style; a | "%Di\n" / trim "\n"
                """
                )
            )
            dict_add(ret, t)

            posts = []
            expr = reliq.expr(
                r"""
                .date td -rowspan l@[1]; a m@E>".+ .+ .+" | "%i" / sed "s/^[^ ]\+ [^ ]\+ //",
                td rowspan l@[1]; {
                    .user font .mediumtxt; * c@[0] | "%Di" trim,
                     div .smalltxt; {
                        .postid.u a name l@[1] | "%(name)v" / sed "s/^pid//",
                        .fields.a * l@[0] | "%i\n" / sed '
                            s/^<a [^>]*><\/a>//
                            s/<br \/>/\n/
                            s/<img src="images[^>]*\/>/*/g
                            s/<br \/>.*<hr \/>/\n/
                            s/<div [^>]*>\(<img [^>]*src="\([^"]*\)"[^>]*\/>\)\?<\/div><br \/>/\2\n/
                            s/<br \/>[^:]*: //
                            s/<br \/>[^:]*: /\n/
                            /<br \/>[^:]*: /!s/<br \/>/\n\0/
                            s/<br \/>[^:]*: \(<div [^>]*><img [^>]* alt="\([^"]*\)"[^>]*\/><\/div><br \/>\([^<]*\)\)\?/\n\3/
                            s/<br \/>[^<]*<br \/>/\n/;
                            s/<br \/>//;
                            s/<strong>[^<]*:<\/strong> //g;
                            s/<strong>//g;
                            s/<\/strong>//g
                            '
                    }
                }
            """
            )

            while True:
                for i in rq.search(
                    r'tr -class bgcolor | "%C\a" / tr "\n\r\t" "   " tr "\a" "\n" sed "N;N;s/\n/\t/g"'
                ).split("\n")[:-1]:
                    tr = i.split("\t")
                    post = {}

                    try:
                        post["body"] = reliq(tr[1]).search(r'td .tablerow | "%i"')
                        post["homepage"] = reliq(tr[2]).search(
                            r'a href title=w>homepage | "%(href)v"'
                        )
                        t = json.loads(reliq(tr[0]).search(expr))
                    except (IndexError, AttributeError):
                        continue

                    for j in ["date", "user", "postid"]:
                        post[j] = t[j]

                    fields = t["fields"]

                    for j, g in enumerate(
                        [
                            "rank",
                            "stars",
                            "avatar",
                            "posts",
                            "registered",
                            "location",
                            "mood",
                        ]
                    ):
                        try:
                            if g == "stars":
                                post[g] = len(fields[j])
                            elif g == "avatar":
                                post[g] = url_merge(ref, fields[j])
                            elif g == "posts":
                                try:
                                    post[g] = int(fields[j])
                                except ValueError:
                                    post[g] = fields[j]
                            else:
                                post[g] = fields[j]
                        except IndexError:
                            pass

                    posts.append(post)

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state, True)

            ret["posts"] = posts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_xmb

        self.trim = True

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.forum_forums_expr = reliq.expr(
            r'font .mediumtxt; a l@[1] href=Ea>"(forumdisplay\.php\?|^[0-9]+/)" | "%(href)v\n"'
        )
        self.forum_threads_expr = reliq.expr(
            r'font .mediumtxt; a href=b>"viewthread.php" l@[1] | "%(href)v\n"'
        )
        self.board_forums_expr = reliq.expr(
            r'a href=Ea>"(forumdisplay\.php\?|^[0-9]+/)" | "%(href)v\n" / sed "s#/./#/#"'
        )
        self.guesslist = [
            {
                "func": "get_thread",
                "exprs": [r"/(.*/)?viewthread.php\?tid=\d+$"],
            },
            {
                "func": "get_forum",
                "exprs": [r"/(.*/)?forumdisplay.php\?fid=\d+$", "/forums?/.+"],
            },
            {
                "func": "get_board",
                "exprs": [r"/(.*/)?index.php\?gid=\d+$"],
            },
            {"func": "get_board", "exprs": None},
        ]

        self.findroot_expr = reliq.expr(
            r'td .nav -style; a href | "%(href)v\n" / sed "/\/$/p" "n" line [-] tr "\n"'
        )
        self.findroot_board = False
        self.findroot_board_expr = None

    def get_next_page(self, rq):
        url = rq.search(
            r'''td .multi; {
                [0] a href rel=next | "%(href)v\n",
                strong; [0] u ssub@; a href | "%(href)v\n"
            } / sed "s/&amp;/\&/g;q" tr "\n"'''
        )
        if not re.search(r"&page=", url):
            return ""
        return url

    def process_board_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
            .categories table C@"td .tablerow l@[2]"; tr -.header l@[1]; {
                .category td .category; * c@[0] | "%Di" trim,
                .category_link [0] a href | "%(href)v",

                .state [0] td l@[1]; img src | "%(src)v",
                [1] td l@[1]; {
                    .description [1:2] font | "%i\a" / sed "/^&nbsp;$/d" "" "\a" tr "\a",
                    a href; {
                        .link [-] * l@[0] | "%(href)v",
                        .name * c@[0] | "%Di" / trim
                    }
                },
                .topics.u [2] td l@[1]; font | "%i",
                .posts.u [3] td l@[1]; font | "%i",
                .lastpost [4] td l@[1]; tr; {
                    [0] td; {
                        .date [0] * C@"a" | "%i" / sed "s#<br />.*##; s/.*>//",
                        a href; {
                            .user * l@[0] | "%Di" / sed "s/.*>//;s/^by //" trim,
                            .user_link * l@[0] | "%(href)v"
                        }
                    },
                    .link [1] td; [0] a href | "%(href)v"
                }
            } | """
            )
        )

        groups = []
        group_name = None
        group_link = None
        group_forums = []
        isfirst = True
        for i in t["categories"]:
            if len(i["category"]):
                if not isfirst:
                    if group_link is None or not group_link.endswith(
                        "misc.php?action=online"
                    ):
                        groups.append(
                            {
                                "name": group_name,
                                "link": group_link,
                                "forums": group_forums,
                            }
                        )
                    group_forums = []
                group_name = i["category"]
                group_link = url_merge(ref, i["category_link"])
                continue

            if len(i["name"]) == 0:
                continue

            i["link"] = url_merge(ref, i["link"])
            i["state"] = url_merge(ref, i["state"])

            lastpost = i["lastpost"]
            if lastpost["user_link"] == lastpost["link"]:
                lastpost["user_link"] = None
            else:
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
            lastpost["link"] = url_merge(ref, lastpost["link"])

            i.pop("category")
            i.pop("category_link")

            group_forums.append(i)

            isfirst = False

        if len(group_forums) != 0 and (
            group_link is None or not group_link.endswith("misc.php?action=online")
        ):
            groups.append(
                {"name": group_name, "link": group_link, "forums": group_forums}
            )

        return {"format_version": "xmb-board", "url": url, "groups": groups}

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                .threads table C@"font .mediumtxt; a href=b>\"viewthread.php?\" l@[1]"; tr -class l@[1]; {
                    .state [0] td l@[1]; img src | "%(src)v",
                    .icon [1] td l@[1]; img src | "%(src)v",
                    [2] td l@[1]; {
                        .sticky.b img | "t",
                        [0] a href; {
                             .link * l@[0] | "%(href)v",
                             .title * l@[0] | "%iD" trim
                        },
                        .lastpage.u u; [-] a | "%i"
                    },
                    [3] td l@[1]; a href; {
                        .user * l@[0] | "%Di" trim,
                        .user_link * l@[0] | "%(href)v"
                    },
                    .replies.u [4] td l@[1]; font | "%i",
                    .views.u [5] td l@[1]; font | "%i",
                    .lastpost [-] td l@[1]; [0] td l@[1:]; {
                        .date [0] * C@"a" | "%i" / sed "s#<br />.*##; s/.*>//",
                        a href; {
                            .user_link * l@[0] | "%(href)v",
                            .user * l@[0] | "%Di" / sed "s/.*>//;s/^by //" trim
                        }
                    }
                } | """
            )
        )

        threads = []

        for i in t["threads"]:
            if len(i["link"]) == 0:
                continue
            if i["lastpost"]["user_link"].startswith("viewthread.php?goto=lastpost&"):
                i["lastpost"]["user_link"] = ""

            dict_url_merge(
                ref,
                i,
                [
                    ["state"],
                    ["icon"],
                    ["link"],
                    ["user_link"],
                    ["lastpost", "user_link"],
                ],
            )
            threads.append(i)

        forums = []

        f = json.loads(
            rq.search(
                r"""
            .forums table C@"td .ctrtablerow l@[2]" C@"font .mediumtxt; a href=b>\"forumdisplay.php?\" l@[1]"; tr -class l@[1]; {
                .state [0] td l@[1]; img src | "%(src)v",
                [1] td l@[1]; {
                    [0] a href; {
                         .link * l@[0] | "%(href)v",
                         .name * l@[0] | "%Di" / trim
                    },
                    .description [-] font | "%i"
                },
                .threads.u [2] td l@[1]; font | "%i",
                .posts.u [3] td l@[1]; font | "%i",
                .lastpost [4] td l@[1]; [0] td l@[1:]; {
                    .date [0] * C@"a" | "%i" / sed "s#<br />.*##; s/.*>//",
                    a href; {
                        .user_link * l@[0] | "%(href)v",
                        .user * l@[0] | "%Di" / sed "s/.*>//;s/^by //" trim
                    }
                }
            } | """
            )
        )

        for i in f["forums"]:
            dict_url_merge(
                ref,
                i,
                [
                    ["state"],
                    ["link"],
                    ["lastpost", "user_link"],
                ],
            )

            forums.append(i)

        return {
            "format_version": "xmb-forum",
            "url": url,
            "forums": forums,
            "threads": threads,
        }



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

from enum import Flag, auto
from importlib.metadata import version

__version__ = version(__package__ or __name__)


class Outputs(Flag):
    write_by_id = auto()
    write_by_hash = auto()
    writers = write_by_id | write_by_hash

    data = auto()

    boards = auto()
    tags = auto()
    forums = auto()
    threads = auto()
    users = auto()
    reactions = auto()

    only_urls_threads = auto()
    only_urls_forums = auto()
    urls = auto()
    save_urls = only_urls_threads | only_urls_forums | urls



import sys

from .enums import Outputs, __version__
from .extractors.extractor import *
from .args import argparser

__all__ = [
    "Outputs",
    "net",
    "ForumExtractor",
    "ForumExtractorIdentify",
    "ItemExtractor",
    "Extractor",
    "invision",
    "phpbb",
    "smf",
    "smf1",
    "smf2",
    "xenforo",
    "xenforo1",
    "xenforo2",
    "xmb",
    "main",
    "__version__",
]


def main():
    args = argparser().parse_args(sys.argv[1:] if sys.argv[1:] else ["-h"])

    disturbed = {"undisturbed": True, "pedantic": False}
    if args.pedantic:
        disturbed = {"undisturbed": False, "pedantic": True}

    output = args.names | Outputs.threads

    if args.nothreads:
        output &= ~Outputs.threads
    if args.users:
        output |= Outputs.users
    if args.reactions:
        output |= Outputs.reactions
    if args.boards:
        output |= Outputs.boards
    if args.tags:
        output |= Outputs.tags
    if args.forums:
        output |= Outputs.forums
    if args.only_urls_threads:
        output = args.names | Outputs.only_urls_threads
    if args.only_urls_forums:
        output = args.names | Outputs.only_urls_forums

    if output == args.names:
        print(
            "Error: Nothing can be downloaded, remove --nothreads option",
            file=sys.stderr,
        )
        return

    headers = {}
    cookies = {}
    if args.cookie is not None:
        for i in args.cookie:
            cookies.update(i)

    if args.header is not None:
        for i in args.header:
            headers.update(i)
        cookie = headers.get("Cookie")
        if cookie is not None:
            headers.pop("Cookie")
            for i in cookie.split(";"):
                pair = i.split("=")
                name = pair[0].strip()
                val = None
                if len(pair) > 1:
                    val = pair[1].strip()
                cookies.update({name: val})

    settings = {
        "output": output,
        "max_workers": args.threads,
        "logger": args.log,
        "failed": args.failed,
        "force": args.force,
        **disturbed,
        "thread_pages_max": args.thread_pages_max,
        "pages_max": args.pages_max,
        "pages_max_depth": args.pages_max_depth,
        "pages_threads_max": args.pages_threads_max,
        "wait": args.wait,
        "wait_random": args.wait_random,
        "retries": args.retries,
        "retry_wait": args.retry_wait,
        "timeout": args.timeout,
        "allow_redirects": args.location,
        "user-agent": args.user_agent,
        "verify": args.insecure,
        "proxies": args.proxies,
        "headers": headers,
        "cookies": cookies,
        "compress_func": args.compression,
    }

    ex = Extractor(**settings)
    func_name = "guess"
    func = getattr(ex, func_name)

    for url in args.urls:
        if isinstance(url, list):
            ex = url[0](**settings)
            func_name = url[1]
            func = getattr(ex, func_name)
            continue

        ret = func(url, **settings)
        if func_name == "findroot":
            print(f"{ret}\t{url}", file=args.output)
        elif func_name == "identify":
            print(
                "{}\t{}".format(ret if ret is None else ret.__class__.__name__, url),
                file=args.output,
            )
        elif args.only_urls_forums:
            for i in ret["urls"]["forums"]:
                print(i, file=args.output)
        elif args.only_urls_threads:
            for i in ret["urls"]["threads"]:
                print(i, file=args.output)

    for i in [args.log, args.failed, args.output]:
        if i not in [sys.stdout, sys.stderr]:
            i.close()



import os
import sys
import argparse
import ast
import gzip
import bz2
import lzma

from .utils import (
    url_valid,
    conv_curl_header_to_requests,
    conv_curl_cookie_to_requests,
)
from .enums import Outputs, __version__
from .extractors.extractor import *


def valid_directory(directory):
    try:
        return os.chdir(directory)
    except:
        raise argparse.ArgumentTypeError(
            'couldn\'t change directory to "{}"'.format(directory)
        )


def valid_names(name):
    if name == "id":
        return Outputs.write_by_id
    elif name == "hash":
        return Outputs.write_by_hash
    else:
        raise KeyError(f'"{name}" is neither id nor hash')


def valid_type(type_name):
    if url_valid(type_name) is not None:
        return type_name

    ret = [Extractor, "guess"]

    forums = {
        "all": Extractor,
        "smf": smf,
        "smf1": smf1,
        "smf2": smf2,
        "phpbb": phpbb,
        "xenforo": xenforo,
        "xenforo1": xenforo1,
        "xenforo2": xenforo2,
        "invision": invision,
        "xmb": xmb,
        "hackernews": hackernews,
        "stackexchange": stackexchange,
        "vbulletin": vbulletin,
    }
    forums_identifiers = ["all", "smf", "xenforo"]
    funcs = ["guess", "identify", "findroot", "thread", "user", "forum", "tag", "board"]

    names = type_name.split(".")
    namesl = len(names)

    if (namesl == 1 and len(names[0]) == 0) or namesl > 2:
        raise KeyError(f"{type_name} is not a valid scraper type")

    if namesl == 2 and len(names[0]) == 0 and len(names[1]) == 0:
        return ret

    if len(names[0]) > 0:
        r = forums.get(names[0])
        if not r:
            raise KeyError(f'"{names[0]}" forum scraper does not exists')
        ret[0] = r

    if namesl == 2 and len(names[1]) > 0:
        if names[1] in funcs:
            if (
                names[1] == "identify"
                and len(names[0])
                and names[0] not in forums_identifiers
            ):
                raise (f'{names[0]} does not have function "identify"')
            if names[1] not in ["guess", "identify", "findroot"]:
                names[1] = "get_" + names[1]
            ret[1] = names[1]
        else:
            raise KeyError(f'"{names[1]}" type does not exists')
    return ret


def valid_compression_type(type_name):
    types = {
        "lzma": lzma.compress,
        "bzip2": bz2.compress,
        "gzip": gzip.compress,
        "none": None,
    }

    if type_name not in types.keys():
        raise KeyError(f"{type_name} is not a valid compression type")
    return types.get(type_name)


def valid_header(src):
    r = conv_curl_header_to_requests(src)
    if r is None:
        raise argparse.ArgumentTypeError('Invalid header "{}"'.format(src))
    return r


def valid_cookie(src):
    r = conv_curl_cookie_to_requests(src)
    if r is None:
        raise argparse.ArgumentTypeError('Invalid cookie "{}"'.format(src))
    return r


def argparser():
    parser = argparse.ArgumentParser(
        description="Forum scraper that aims to be an universal, automatic and extensive.",
        add_help=False,
    )

    parser.add_argument(
        "urls",
        metavar="URL|TYPE",
        type=valid_type,
        nargs="*",
        help="url pointing to source. Specifying type changes scraper for all next urls. Type consists of forum name which can be all, invision, phpbb, smf, smf1, smf2, xenforo, xenforo1, xenforo2, xmb, stackexchange, hackernews and it can by followed by a '.' and function name which can be guess, thread, forum, tag, board, user, identify, findroot. By default set to all.guess equivalent to .guess and .",
    )

    general = parser.add_argument_group("General")
    general.add_argument(
        "-h",
        "--help",
        action="help",
        help="Show this help message and exit",
    )
    general.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
        help="Print program version and exit",
    )
    general.add_argument(
        "-t",
        "--threads",
        metavar="NUM",
        type=int,
        help="run tasks using NUM of threads",
        default=1,
    )
    general.add_argument(
        "-p",
        "--pedantic",
        action="store_true",
        help="Exit if anything fails",
    )

    files = parser.add_argument_group("Files")
    files.add_argument(
        "-d",
        "--directory",
        metavar="DIR",
        type=valid_directory,
        help="Change directory to DIR",
    )
    files.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="forcefully overwrite files, may cause redownloading if found urls are not unique",
    )
    files.add_argument(
        "--names",
        metavar="NAME",
        type=valid_names,
        help="Change naming convention of created files to NAME, which can be either id or hash",
        default=Outputs.write_by_id,
    )
    files.add_argument(
        "-L",
        "--location",
        action="store_true",
        help="Allow for redirections, can be dangerous if credentials are passed in headers",
    )
    files.add_argument(
        "--compression",
        metavar="ALGO",
        type=valid_compression_type,
        help="Compress created files using set algorithm, that can be none, gzip, bzip2, lzma",
        default=None,
    )
    files.add_argument(
        "-l",
        "--log",
        metavar="FILE",
        type=lambda x: open(x, "w"),
        help="log results to FILE (by default set to stdout)",
        default=sys.stdout,
    )
    files.add_argument(
        "-F",
        "--failed",
        metavar="FILE",
        type=lambda x: open(x, "w"),
        help="log failures to FILE (by default set to stderr)",
        default=sys.stderr,
    )
    files.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        type=lambda x: open(x, "w"),
        help="store results to FILE (by default set to stdout), work only for identify and findroot functions and --only-urls options",
        default=sys.stdout,
    )

    settings = parser.add_argument_group("Settings")
    settings.add_argument(
        "--nothreads",
        action="store_true",
        help="do not download threads (works only when page passed is not a thread)",
    )
    settings.add_argument(
        "--users",
        action="store_true",
        help="download users",
    )
    settings.add_argument(
        "--reactions",
        action="store_true",
        help="download reactions",
    )
    settings.add_argument(
        "--boards",
        action="store_true",
        help="create boards files",
    )
    settings.add_argument(
        "--tags",
        action="store_true",
        help="create tags files",
    )
    settings.add_argument(
        "--forums",
        action="store_true",
        help="create forums files",
    )
    settings.add_argument(
        "--only-urls-forums",
        action="store_true",
        help="output urls to found forums to output file without scraping",
    )
    settings.add_argument(
        "--only-urls-threads",
        action="store_true",
        help="output urls to found threads to output file without scraping",
    )
    settings.add_argument(
        "--thread-pages-max",
        metavar="NUM",
        type=int,
        help="set max number of pages traversed in threads",
    )
    settings.add_argument(
        "--pages-max",
        metavar="NUM",
        type=int,
        help="set max number of pages traversed in pages amassing threads e.g. forums, tags, boards",
    )
    settings.add_argument(
        "--pages-max-depth",
        metavar="NUM",
        type=int,
        help="set max recursion depth",
    )
    settings.add_argument(
        "--pages-forums-max",
        metavar="NUM",
        type=int,
        help="set max number of forums to be processed in every page",
    )
    settings.add_argument(
        "--pages-threads-max",
        metavar="NUM",
        type=int,
        help="set max number of threads to be processed in every page",
    )

    request_set = parser.add_argument_group("Request settings")
    request_set.add_argument(
        "-w",
        "--wait",
        metavar="SECONDS",
        type=float,
        help="Sets waiting time for each request to SECONDS",
    )
    request_set.add_argument(
        "-W",
        "--wait-random",
        metavar="MILISECONDS",
        type=int,
        help="Sets random waiting time for each request to be at max MILISECONDS",
    )
    request_set.add_argument(
        "-r",
        "--retries",
        metavar="NUM",
        type=int,
        help="Sets number of retries for failed request to NUM",
    )
    request_set.add_argument(
        "--retry-wait",
        metavar="SECONDS",
        type=float,
        help="Sets interval between each retry",
    )
    request_set.add_argument(
        "-m",
        "--timeout",
        metavar="SECONDS",
        type=float,
        help="Sets request timeout",
    )
    request_set.add_argument(
        "-k",
        "--insecure",
        action="store_false",
        help="Ignore ssl errors",
    )
    request_set.add_argument(
        "-A",
        "--user-agent",
        metavar="UA",
        type=str,
        help="Sets custom user agent",
    )
    request_set.add_argument(
        "-x",
        "--proxies",
        metavar="DICT",
        type=lambda x: dict(ast.literal_eval(x)),
        help='Set requests proxies dictionary, e.g. -x \'{"http":"127.0.0.1:8080","ftp":"0.0.0.0"}\'',
    )
    request_set.add_argument(
        "-H",
        "--header",
        metavar="HEADER",
        type=valid_header,
        action="append",
        help="Set header, can be used multiple times e.g. -H 'User: Admin' -H 'Pass: 12345'",
    )
    request_set.add_argument(
        "-b",
        "--cookie",
        metavar="COOKIE",
        type=valid_cookie,
        action="append",
        help="Set cookie, can be used multiple times e.g. -b 'auth=8f82ab' -b 'PHPSESSID=qw3r8an829'",
    )

    return parser



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..utils import dict_add, url_merge_r, url_merge
from .common import ItemExtractor, ForumExtractor
from .identify import identify_phpbb


class phpbb(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    r"/((.*)?viewtopic\.php(.*)[\&\?]t=|.*-t)(\d+)",
                    4,
                )
            ]
            self.trim = True

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "phpbb-2+-thread", "url": url, "id": int(i_id)}
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                .title {
                    div #page-body; B>h[1-6] [0]; [0] a href=b>"./viewtopic.php" | "%Di" trim ||
                    B>h[1-6]; [0] a href=b>"./viewtopic.php" | "%Di" trim
                },
                .path.a {
                    [0] ul #nav-breadcrumbs; a; span itemprop | "%Di\n",
                    div #page-header; li .icon-home; a | "%Di\n"
                } /  trim "\n"
            """
                )
            )
            dict_add(ret, t)

            posts = []
            expr = reliq.expr(
                r"""
                .posts div #B>"p[0-9]*"; {
                    .postid.u div #B>"p[0-9]*" l@[0] | "%(id)v" / sed "s/^p//",
                    .date p .author | "%i" / sed "s/<\/time>$//;s/.*>//; s/.*;//; s/^ *//; s/^([a-z]* )+//; /^$/d" "E",
                    .content div .content | "%i",
                    .signature div .signature #B>sig[0-9]* | "%i",
                    dl .postprofile #B>profile[0-9]*; {
                        dt l@[1]; {
                            .avatar img src | "%(src)v",
                            .user a c@[0] | "%Di" trim,
                            .userid.u a href c@[0] | "%(href)v" / sed "s/.*[&;]u=([0-9]+).*/\1/" "E",
                        },
                        .userinfo_temp.a("\a") dd l@[1] m@vf>"&nbsp;" | "%i\a" / tr '\n\t' sed "s/<strong>([^<]*)<\/strong>/\1/g; s/ +:/:/; /<ul [^>]*class=\"profile-icons\">/{s/.*<a href=\"([^\"]*)\" title=\"Site [^\"]*\".*/Site\t\1/;t;d}; /^[^<>]+:/!{s/^/Rank:/};s/: */\t/" "E" "\a"
                    }
                } |
            """
            )

            while True:
                t = json.loads(rq.search(expr))
                posts += t["posts"]

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state)

            for i in posts:
                i["avatar"] = url_merge_r(ref, i["avatar"])
                i["userinfo"] = []
                for j in i["userinfo_temp"]:
                    t = j.split("\t")
                    t_len = len(t)
                    if t_len > 2 or (t_len == 0 and len(t[0]) == 0):
                        continue
                    if t_len < 2:
                        t.append("")
                    i["userinfo"].append({"key": t[0], "value": t[1]})
                i.pop("userinfo_temp")

            ret["posts"] = posts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_phpbb

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.forum_forums_expr = reliq.expr(
            r'li; a .forumtitle href | "%(href)v\n" / sed "s/^\.\///;s/&amp;/\&/g"'
        )
        self.forum_threads_expr = reliq.expr(
            r'li; a .topictitle href | "%(href)v\n" / sed "s/^\.\///;s/&amp;/\&/g"'
        )
        self.board_forums_expr = self.forum_forums_expr
        self.guesslist = [
            {
                "func": "get_thread",
                "exprs": [r"/(.*/)?viewtopic.php.*[\&\?]t=\d+"],
            },
            {
                "func": "get_forum",
                "exprs": [r"/(.*/)?viewforum.php"],
            },
            {
                "func": "get_board",
                "exprs": [r"/(.*/)?index.php"],
            },
            {"func": "get_board", "exprs": None},
        ]

        self.findroot_expr = reliq.expr(
            r"""
                {
                    div .nav-tabs; li .tab .forums; a .nav-link,
                    div #page-footer,
                    div .f_footer,
                }; [0] a href | "%(href)v\n" / line [0] sed "s/&amp;/\&/g" tr "\n"
            """
        )
        self.findroot_board = True
        self.findroot_board_expr = re.compile(
            r"^(/[^\.-])?/((board|forum|foro)s?|index\.(php|html)|community|communaute|comunidad)(/|\?[^/]*)?$",
        )

    def get_next_page(self, rq):
        url = rq.search(r'[0] a href rel=next | "%(href)v" / sed "s/&amp;/\&/g;q"')
        if len(url) == 0:
            url = rq.search(
                r'div .topic-actions; div .pagination l@[1]; span l@[1]; E>(a|strong) -href=# | "%(href)v\n" / sed "/^$/{N;s/\n//;s/&amp;/\&/g;p;q}" "n"'
            )[:-1]

        if not re.search(r"&start=[0-9]+$", url):
            return ""
        return url

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                * #page-body; {
                    .categories div .forabg; {
                        [0] ul -.forums .topiclist; dt; { a, * self@ }; [0] * self@; {
                            .name * self@ | "%Dt" / trim,
                            .link * self@ | "%(href)v"
                        },
                        .forums ul .forums .topiclist; li child@; dl child@; {
                            .state {
                                * title self@ | "%(title)v\a",
                                * style=b>background-image: self@ | "%(style)v" sed "s/.*url(//;s/);.*//; s/.*\///; s/\..*//",
                                * class self@ | "%(class)v" tr ' ' '\n' sed '/^icon$/d; /^row-item$/d; /^elegant-row$/d'
                            } / sed "s/\a.*//" trim,
                            dt child@; {
                                [0] a href; {
                                    .name * self@ | "%Di" / trim,
                                    .link * self@ | "%(href)v"
                                },
                                .icon span .forum-image; img src | "%(src)v",
                                .description * self@ | "%i" tr "\n\t\r" sed "/<br *\/>/!d; s/.*-->//; s/[^<]*(<br *\/?>(.*)|<)/\2/g; s/<strong>.*//; s/<br *\/?>/\n/g; s/<span[^>]*>([^<]*)<\/span>/\1/g; s/<div.*//" "E" trim,
                                .childboards a .subforum; {
                                    .state * self@ | "%(title)v",
                                    .name * self@ | "%Dt" / trim,
                                    .link * self@ | "%(href)v"
                                } | ,
                                .moderators strong c@[0] m@b>Mod m@e>: l@[1:2]; a -.subforum ssub@; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                } | ,
                                div .forum-statistics; {
                                    .topics2.u [0] span .value | "%i",
                                    .posts2.u [1] span .value | "%i",
                                }
                            },
                            .topics.u dd .b>topics child@ | "%t" tr " ,",
                            .posts.u dd .b>posts child@ | "%t" tr " ,",
                            .redirects.u dd .redirect | "%T",
                            .lastpost dd .b>lastpost child@; span child@; {
                                a ( .b>username )( -class ) c@[0]; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                },
                                [0] a ( .lastsubject )( c@[1:] ); {
                                    .title * self@ | "%(title)Dv" / trim,
                                    .link * self@ | "%(href)v"
                                },
                                .date {
                                    time datetime | "%(datetime)v\a",
                                    * self@ | "%i" tr "\n\t\r" sed "/<br/!d; s/.*<br *\/?>//; s/^on //; s/&nbsp;//g" "E"
                                } / sed "s/\a.*//" trim
                            }
                        } |
                    } | ,
                    .threads div .forumbg; {
                        .name ul .topiclist -.topics; [0] dt | "%DT" / trim,
                        .threads ul .topiclist .topics; li child@; dl child@; {
                            .state {
                                * title self@ | "%(title)v\a",
                                * style=b>background-image: self@ | "%(style)v" sed "s/.*url(//;s/);.*//; s/.*\///; s/\..*//",
                                * class self@ | "%(class)v" tr ' ' '\n' sed '/^icon$/d; /^row-item$/d; /^elegant-row$/d'
                            } / sed "s/\a.*//" trim,
                            dt child@; {
                                [0] a href; {
                                    .title * self@ | "%Di" / trim,
                                    .link * self@ | "%(href)v"
                                },
                                .lastpage.u * .pagination; [-] a c@[0] m@B>"[0-9]" | "%i",
                                [-] a href c@[0]; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                },
                                .date {
                                    time datetime | "%(datetime)v\a",
                                    {
                                        * self@ | "%i",
                                        div .topic-poster | "%t"
                                    } / tr "\n\t\r" sed "s/.*>//; s/^on //; s/&nbsp;//g; s/.*&raquo;//" "E"
                                } / sed "s/\a.*//" trim
                            },
                            .posts.u dd .posts | "%i",
                            .views.u dd .views | "%i",
                            .lastpost dd .b>lastpost child@; span child@; {
                                a ( .b>username )( -class ) c@[0]; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                },
                                .link [0] a ( .lastsubject )( c@[1:] ) | "%(href)v",
                                .date {
                                    time datetime | "%(datetime)v\a",
                                    * self@ | "%i" tr "\n\t\r" sed "/<br/!d; s/.*<br *\/?>//; s/^on //; s/&nbsp;//g" "E"
                                } / sed "s/\a.*//" trim
                            }
                        } |
                    } |
                }
                """
            )
        )

        categories = t["categories"]

        for i in t["categories"]:
            i["link"] = url_merge(ref, i["link"])

            for j in i["forums"]:
                for g in j["childboards"]:
                    g["link"] = url_merge(ref, g["link"])

                j["link"] = url_merge(ref, j["link"])
                for g in j["moderators"]:
                    g["user_link"] = url_merge(ref, g["user_link"])

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])

                if j["posts"] == 0:
                    j["posts"] = j["posts2"]
                if j["topics"] == 0:
                    j["topics"] = j["topics"]
                j.pop("posts2")
                j.pop("topics2")

                j["icon"] = url_merge(ref, j["icon"])

        threads = t["threads"]

        for i in threads:
            for j in i["threads"]:
                j["link"] = url_merge(ref, j["link"])
                j["user_link"] = url_merge(ref, j["user_link"])

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])

        return {
            "format_version": "phpbb-2+-forum",
            "url": url,
            "categories": categories,
            "threads": threads,
        }



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..utils import dict_add, url_merge_r, url_merge
from .common import ItemExtractor, ForumExtractor, ForumExtractorIdentify
from .identify import identify_smf1, identify_smf2


guesslist = [
    {
        "func": "get_thread",
        "exprs": [r"/.*([?/&;]topic[=,]|-t)+(\d+)"],
    },
    {
        "func": "get_forum",
        "exprs": [r"/.*([?/&;]board[=,]|-t)+(\d+)"],
    },
    {
        "func": "get_board",
        "exprs": [r"/(.*/)?index.php"],
    },
    {"func": "get_board", "exprs": None},
]


class smf1(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/.*([?/&;]topic[=,]|-t)(\d+)"),
                    2,
                )
            ]

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "smf-1-thread", "url": url, "id": int(i_id)}
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                .title * #top_subject | "%i" / sed "s/[^:]*: //;s/([^(]*)//;s/&nbsp;//g;s/ *$//; s/<[^>]*>//g; s/^Topic: //" decode trim,
                .viewed.u {
                    td #top_subject | "%i" /  sed "s/.*\(//g;s/.* ([0-9]+) .*/\1/" "E" ||
                    * #topic_header_views | "%i"
                },
                .path.a("\n") {
                    div a@[0]; E>(div|span) .nav l@[1]; a .nav | "%i\n" ||
                    [0] * .linktree; a child@ | "%i\n"
                } / line [:-1] decode trim "\n"
            """
                )
            )
            dict_add(ret, t)

            posts = []
            expr = reliq.expr(
                r"""
                .posts form #quickModForm; {
                    table l@[1]; tr l@[:2] m@B>"id=\"subject_[0-9]*\""; tr l@[0] m@v>".googlesyndication.com/" ||
                    ul child@; li -.pageindex_li child@
                }; {
                    .postid.u [0] * #B>subject_[0-9]* | "%(id)v" / sed "s/.*_//",
                    .date {
                        td valign=middle; div .smalltext | "%i" / sed "s/.* ://;s/^<\/b> //;s/ &#187;//g;s/<br *\/>.*//;s/<[^>]*>//g;s/ *$//" ||
                        [0] *  #b>subject_; [0] span c@[0] | "%i" / sed "s/^- //" trim
                    },
                    .body div .post | "%i",
                    .signature div .signature | "%i",
                     td valign=top rowspan=2; {
                        b l@[1]; [0] a href; {
                            .user * self@ | "%Di" / trim,
                            .userid.u * self@ | "%(href)v" / sed "s/.*;//;s/.*-//;s/\.html$//;s/^u//;s/^=//"
                        },
                        .avatar [0] img .avatar src | "%(src)v"
                    },
                    .edited td #B>modified_[0-9]*; * c@[0] | "%i",
                    .score span #B>gpbp_score_[0-9]* | "%i",
                    .attachments.a("\t") a href #B>link_[0-9]* | "%(href)v\t",
                    .userinfo.a("\n") {
                        td valign=top rowspan=2; div .smalltext | "%i\n" / sed "s/\(<br \/>\)\+/\n/g;s/\t//g" sed "/^$/d;/<img.* class=\"avatar\"/d" trim "\n" sed "s/< *br *\/ *>/ /g;s/< *br *>/ /g; /^ *$/d" trim "\n" ||
                        ul .tabs_list; li child@ | "%(class)v: %i" tr "\n\r" trim echo "" "\n"
                    }
                } |
            """
            )

            while True:
                t = json.loads(rq.search(expr))
                for i in t["posts"]:
                    i["avatar"] = url_merge_r(ref, i["avatar"])
                posts += t["posts"]

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state)

            for i in posts:
                ui = []
                for j in i["userinfo"]:
                    r = re.search("(^[A-Za-z0-9_-]+): (.*)$", j)
                    if r is not None:
                        key = r[1]
                        value = r[2]
                        if key == "poster_avatar" and len(i["avatar"]) == 0:
                            navatar = reliq(value).search('[0] img src | "%(src)v"')
                            i["avatar"] = url_merge(ref, navatar)
                            continue
                        elif key == "dropmenu" and len(i["user"]) == 0:
                            i["user"] = reliq(value).search(
                                '[0] * c@[0] m@>[1:] | "%Di" trim'
                            )
                            continue

                    ui.append(j)
                i["userinfo"] = ui

            ret["posts"] = posts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_smf1

        self.trim = True

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.forum_forums_expr = reliq.expr(
            r'td .windowbg2 m@B>"name=\"b[0-9]*\""; b l@[1]; a href l@[1] | "%(href)v\n"'
        )
        self.forum_threads_expr = reliq.expr(
            r'td .B>"windowbg[0-9]*" m@"<span class=\"smalltext\""; a href l@[1] | "%(href)v\n" / sed "s/[.;]msg[^\/]*#new$//;s/#new$//"'
        )
        self.board_forums_expr = self.forum_forums_expr
        self.guesslist = guesslist

        self.findroot_expr = reliq.expr(
            r"""
            {
                * .E>"maintab(_active)?_back",
                div #toolbar; div .tabs; * role=menuitem; a l@[1],
                div #headerarea; a l@[1],
                div #E>(myNavbar|toolbar[0-9]|topmenu|menu|nav); li,
                body; [0] * l@[1]; table l@[0]; tr .B>windowbg[123]; td; a l@[1] C@"img" c@[1],
            }; a href | "%(href)v\n" / sed "
                \#(/((board|forum|foro)s?|community)(/(index\.(html|php))?|\.(html|php))?(\?[^/]*)?|[;&?]action=forum)$#{p;q}
                1!G; h
                $p
            " "En" line [-] sed "s/&amp;/\&/g" tr "\n"
            """
        )
        self.findroot_board = True
        self.findroot_board_expr = re.compile(
            r"^(/[^\.-])?/((forum|foro|board)s?|index\.(php|html)|community|communaute|comunidad)(/|\?[^/]*)?$",
        )

    def get_next_page(self, rq):
        return rq.search(
            r"""
            b m@B>"[0-9]"; [0] a .navPages href ssub@ | "%(href)v" ||
            [0] * .current_page; [0] a ssub@; * c@[0] self@ | "%(href)v" / sed "s/?PHPSESSID=[^\/&]*&amp;/?/"
        """
        )

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                .categories {
                    div #bodyarea; div .tborder style l@[1],
                    div #mainarea; div -class l@[1],
                    [0] table .bordercolor C@"tr .titlebg"
                }; {
                    .name * .E>catbgf?; [0] * c@[0] m@>[1:] | "%Di" trim,
                    .forums [0] table l@[:2]; tr l@[:2]; {
                        .childboards [0] td l@[1]; * l@[0] -C@"img"; a href; {
                            .link * l@[0] | "%(href)v",
                            .name * l@[0] | "%Di" / trim,
                            .state * l@[0] | "%(title)v" sed "s/ (.*//",
                            .topics.u * l@[0] | "%(title)v" sed "s/.* (//; s/)$//; s/,.*//; s/.*: //",
                            .posts.u * l@[0] | "%(title)v" sed "s/.* (//; s/)$//; s/.*, //; s/.*: //"
                        } | ,

                        .state [0] td l@[1]; img title | "%(title)v",
                        [1] td l@[1]; {
                            [0] a href; {
                                .link * l@[0] | "%(href)v",
                                .name * l@[0] | "%Di" / trim
                            },
                            .description * l@[0] | "%t" trim sed "s/^&nbsp;//" trim,
                            .moderators div; i; a href; {
                                .user_link * l@[0] | "%(href)v",
                                .user * l@[0] | "%Di" trim
                            } |
                        },
                        [2] td l@[1]; {
                            {
                                span,
                                small,
                                * l@[0]  c@[1:]
                            }; {
                                .posts.u * l@[0] | "%t" line [0],
                                .topics.u * l@[0] | "%t" line [1:] " "
                            },
                            .posts2.u * l@[0] c@[0] | "%i"
                        },

                        .topics2.u [3] td l@[1]; {
                            span .largetext l@[1] | "%i",
                            * l@[0] c@[0] | "%i"
                        },
                        .lastpost {
                            [3:4] td l@[1]; {
                                [0] span l@[1]; * .smalltext l@[0],
                                [0] small l@[1]
                            },
                            [1] td l@[1]; span .smalltext
                        }; {
                            [0] a href=aE>(action=profile|/profiles/); {
                                .user_link * l@[0] | "%(href)v",
                                .user [0] * l@[0] | "%Di" trim
                            },
                            [0] a href -href=aE>(action=profile|/profiles/); {
                                .link * l@[0] | "%(href)v",
                                .title * l@[0] | "%Di" / trim
                            },
                            .date * l@[0] | "%i" / tr "\n\r" sed "/^</!s/<.*//; s/.*<br \/>//; s/.*<\/a>//; s/<[^>]*>//g; s/\t//g; s/^on //; s/^  *//;"
                        }
                    } |
                } | ,
                .categories2 div #bodyarea_inner; div .boardindex l@[1]; {
                    .name div .cat_bar l@[1]; h3 | "%Di" / trim,
                    .forums li l@[2]; {
                        .childboards div .childboards; a; {
                            .link * l@[0] | "%(href)v",
                            .name * l@[0] | "%Di" / trim,
                            .state * l@[0] | "%(title)v" sed "s/ (.*//",
                            .topics.u * l@[0] | "%(title)v" sed "s/.* (//; s/)$//; s/,.*//; s/.*: //",
                            .posts.u * l@[0] | "%(title)v" sed "s/.* (//; s/)$//; s/.*, //; s/.*: //"
                        } | ,
                        .state a .board_icon; img title | "%(title)v",
                        div .info; {
                            h4; a; {
                                .link * l@[0] | "%(href)v",
                                .name * l@[0] | "%Di" / trim,
                                .posts.u * l@[0] | "%(title)v",
                                .topics.u * l@[0] | "%(title)v" line [1:] " "
                            },
                            .description p l@[1] | "%i"
                        },
                        .lastpost div .lastpost; {
                            a href; {
                                .link * l@[0] | "%(href)v",
                                .title * l@[0] | "%Di" / trim
                            },
                            .date * l@[0] | "%i" sed "s/<.*//; s/.*: //;q" trim,
                            [-] a href=a>"action=profile"; {
                                .user [0] * l@[0] | "%Di" trim,
                                .user_link * l@[0] | "%(href)v"
                            },
                            .user2 * l@[0] | "%i" tr "\n" sed "s/.*>//; s/.*&nbsp;//" decode trim,
                        }
                    } |
                } | ,
                .threads [0] table .bordercolor C@"tr l@[1:2]; td [2] l@[1]; a href=aE>\"([&?;/]topic[=,]|-t)[0-9]+\.0\""; [1:] tr l@[1:2]; {
                    .type1 [0] td l@[1]; img src | "%(src)v",
                    .type2 [1] td l@[1]; img src | "%(src)v",
                    [2] td l@[1]; {
                        .icons.a img src | "%(src)v\n",
                        [0] a href; {
                            .link * l@[0] | "%(href)v",
                            .title * l@[0] | "%Di" / trim
                        },
                        .lastpage.u {
                            span l@[1],
                            small l@[1]
                        }; [1:] a; [-] a -m@f>"All" | "%i"
                    },
                    [3] td l@[1]; a href; {
                        .user_link * l@[0] | "%(href)v",
                        .user [0] * l@[0] | "%Di" trim
                    },
                    .replies.u [4] td l@[1] | "%T",
                    .views.u [5] td l@[1] | "%T",
                    .lastpost [6] td l@[1]; span .smalltext l@[1]; {
                        [0] a href; {
                            .user_link * l@[0] | "%(href)v",
                            .user [0] * l@[0] | "%i" trim
                        },
                        .date * l@[0] | "%i" sed "s/<br \/>.*//; s/<[^>]*>//g;q" trim
                    }
                } | ,
                .threads2 div #messageindex; li l@[2] -.pageindex_li c@[!0]; {
                    div .info; {
                        span #b>msg_; a; {
                            .link * l@[0] | "%(href)v",
                            .name * l@[0] | "%Di" / trim,
                            .replies.u * l@[0] | "%(title)v",
                            .views.u * l@[0] | "%(title)v" line [1:] " "
                        },
                        [-] a l@[1]; {
                            .user_link * l@[0] | "%(href)v",
                            .user [0] * l@[0] | "%Di" trim
                        },
                        .lastpage.u span #b>pages; [-] a | "%i",
                        .icons.a img l@[1] | "%(src)v\n"
                    },
                    .lastpost div .lastpost; {
                        a href c@[0]; {
                            .user [0] * l@[0] | "%Di" trim,
                            .user_link * l@[0] | "%(href)v"
                        },
                        .date * l@[0] | "%i" tr "\n\t" sed "s/.*>//; s/.*&nbsp;//"
                    }
                } |
                """
            )
        )

        categories = t["categories"]
        categories_forums = []

        if len(categories) == 0:
            categories = t["categories2"]

        for i in categories:
            prev_index = None
            for index, j in enumerate(i["forums"]):
                for g in j["childboards"]:
                    g["link"] = url_merge(ref, g["link"])

                if len(j["link"]) == 0 or j["link"].find(";sort=") != -1:
                    if prev_index is not None and len(j["childboards"]) != 0:
                        categories_forums[prev_index]["childboards"] += j["childboards"]
                    continue

                j["link"] = url_merge(ref, j["link"])
                try:
                    if j["posts"] == 0:
                        j["posts"] = j["posts2"]
                    j.pop("posts2")
                except KeyError:
                    pass

                try:
                    if j["topics"] == 0:
                        j["topics"] = j["topics2"]
                    j.pop("topics2")
                except KeyError:
                    pass

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
                if len(lastpost["user"]) == 0:
                    lastpost["user"] = lastpost.get("user2", "")
                try:
                    lastpost.pop("user2")
                except KeyError:
                    pass

                if j.get("moderators") is None:
                    j["moderators"] = []

                for g in j["moderators"]:
                    g["user_link"] = url_merge(ref, g["user_link"])

                prev_index = len(categories_forums)
                categories_forums.append(j)

            i["forums"] = categories_forums

        threads = t["threads"]
        if len(threads) == 0:
            threads = t["threads2"]
        outthreads = []

        for i in threads:
            if len(i["link"]) == 0:
                continue

            i["link"] = url_merge(ref, i["link"])
            i["user_link"] = url_merge(ref, i["user_link"])
            i["lastpost"]["user_link"] = url_merge(ref, i["lastpost"]["user_link"])
            i["type1"] = url_merge_r(ref, i.get("type1", ""))
            i["type2"] = url_merge_r(ref, i.get("type2", ""))

            icons = i["icons"]
            for j, g in enumerate(icons):
                icons[j] = url_merge(ref, g)

            outthreads.append(i)

        return {
            "format_version": "smf-1-forum",
            "url": url,
            "categories": categories,
            "threads": outthreads,
        }


class smf2(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/.*([?/&;]topic[=,])(\d+)"),
                    2,
                )
            ]
            self.trim = True

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "smf-2-thread", "url": url, "id": int(i_id)}
            page = 0

            forumposts = rq.filter(r"div #forumposts")
            title = forumposts.search(
                r'div .cat_bar; h3 .catbg | "%i\n" / sed "s/<[^>]*>//g;s/ &nbsp;/ /;s/ ([^)]*)$//;s/^[^:]*: //" decode'
            )[:-1]
            if len(title) > 0:
                viewed = forumposts.search(
                    r'div .cat_bar; h3 .catbg | "%i\n" / sed "s/<[^>]*>//g;s/ &nbsp;/ /;s/.* (\([^)]*\))$/\1/;s/.* \([0-9]*\) .*/\1/"'
                )[:-1]
            else:
                title = forumposts.search(r'h1 | "%i"') + rq.search(
                    r'B>h[0-9] .display_title; span #top_subject | "%Di"'
                )
                viewed = forumposts.search(
                    r'div .display-info;  li m@v>"comments" | "%i\n" / sed "s/<[^>]*>//g; s/ .*//"'
                )[:-1]

            ret["title"] = title
            try:
                ret["viewed"] = int(viewed)
            except ValueError:
                ret["viewed"] = viewed

            ret["path"] = json.loads(
                rq.search(
                    r"""
                .path.a {
                    div .navigate_section [0]; li; a -href=a>action= l@[1]; * c@[0] | "%Di\n" / line [:-1],
                    div .container; ol .breadcrumb [0]; li l@[1]; a l@[1]; * c@[0] | "%Di\n" / line [:-1]
                } / trim "\n"
            """
                )
            )["path"]

            posts = []
            expr = reliq.expr(
                r"""
                .posts form #quickModForm; div l@[1]; {
                    .postid.u div #B>msg_[0-9]* | "%(id)v" / sed "s/^msg_//",
                    .date div .postarea; div .keyinfo; E>(div|a) .smalltext | "%i" / sed "s/.*>//;s/^ //;s/&#187;//;s/ $//",
                    .body div .post; div #B>msg_[0-9]* l@[1] | "%i",
                    .signature div .signature | "%i",
                    .edited * #B>modified_[0-9]* | "%i",
                    .attachments.a div .attached; div .attachments_top; a href | "%(href)v\n",
                    .user div .poster; h4; [0] a l@[1] | "%Di" trim,
                    .userid.u div .poster; h4; a href l@[1] | "%(href)v" / sed "s/^.*;u=//",
                    .avatar div .poster; { ul #B>msg_[0-9]*_extra_info, ul .user_info }; li .avatar; img src | "%(src)v",
                    .userinfo div .poster; { ul #B>msg_[0-9]*_extra_info, ul .user_info }; li -.avatar class; {
                        .key * l@[0] | "%(class)v",
                        .value * l@[0] | "%i"
                    } |
                } |
            """
            )

            while True:
                t = json.loads(rq.search(expr))
                outt = []
                for i in t["posts"]:
                    if i["postid"] == 0 and i["date"] == "" and i["body"] == "":
                        continue
                    i["avatar"] = url_merge_r(ref, i["avatar"])
                    outt.append(i)

                posts += outt

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state, self.trim)

            ret["posts"] = posts
            return ret

        def get_improper_url(self, url, rq, settings, state):
            if rq is None:
                rq, ref = self.session.get_html(url, settings, state, self.trim)

            try:
                i_id = str(int(rq.search('input name=sd_topic value | "%(value)v"')))
            except ValueError:
                print(
                    'url leads to improper forum - "{}"'.format(url),
                    file=settings["logger"],
                )
                return [None, 0]

            return [rq, i_id]

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_smf2

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.forum_forums_expr = reliq.expr(
            r'{ * #E>board_[0-9]+, td .windowbg2 }; a E>(name|id)=E>b[0-9]+ href | "%(href)v\n"'
        )
        self.forum_threads_expr = reliq.expr(
            r'span #B>msg_[0-9]*; a href | "%(href)v\n"'
        )
        self.board_forums_expr = self.forum_forums_expr
        self.guesslist = guesslist

        self.findroot_expr = reliq.expr(
            r"""
            {
                * #E>(menu|site_menu|main_menu|nav|navigation|navigation-root|header|kwick|button_home|toolbar-l); li,
                div #submenu,
                ul .E>(si[td]ebar-menu|main-navigation|topnav),
            }; a href | "%(href)v\n" / sed "
                /[;&?]action=/{/=forum$/!bskip}
                \#(/((board|forum|foro)s?|community)(\.([a-zA-Z0-9_-]+\.)+[a-zA-Z]+)?(/(index\.(html|php))?|\.(html|php))?(\?[^/]*)?|[;&?]action=forum)$#{p;q}
                :skip
                1!G; h
                $p
            " "En" line [-] sed "s/&amp;/\&/g" tr "\n"
            """
        )
        self.findroot_board = True
        self.findroot_board_expr = re.compile(
            r"^(/[^\.-])?/((board|forum|foro)s?|index\.(php|html)|community|communaute|comunidad)(/|\?[^/]*)?$",
        )

    def get_next_page(self, rq):
        return rq.search(
            r'div .pagelinks [0]; E>(a|span|strong) m@vB>"[a-zA-Z .]" l@[1] | "%(href)v %i\n" / sed "$q; /^ /{N;D;s/ .*//;p;q}" "n"'
        )[:-1]

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                    .categories {
                        div #boardindex_table; * #b>category_,
                        div #bodyarea; div .categoryframe child@,
                        div #E>board_[0-9]+_childboards,
                        div #childboards; table .boardsframe child@
                    }; {
                        {
                            * #E>category_[0-9]+,
                            div .categoryframe self@,
                            div #E>board_[0-9]+_childboards self@,
                            table .boardsframe self@; [0] tr child@
                        }; [0] ( h3 )( td ) .catbg; {
                            .name * self@ | "%Dt" trim,
                            .description [0] div .desc ssub@ | "%i"
                        },
                        .forums {
                            {
                                * #E>category_[0-9]+_boards,
                                div .categoryframe self@; table .boardsframe child@
                            }; ( tr )( div ) -style child@,
                            table .boardsframe self@; [1:] tr child@,
                            div #E>board_[0-9]+_childboards self@; * #E>board_[0-9]+,
                        }; {
                            .status [0] * ( .b>icon )( .board_icon ); [0] ( a )( img )( span ) title | "%(title)v",
                            * .info child@; {
                                [0] a href; {
                                    .link * self@ | "%(href)v",
                                    .name * self@ | "%Di" trim
                                },
                                .description [0] ( p -id -.moderators )( div .board_description ) child@ | "%i",
                            },
                            .childboards * .children; a; {
                                .link * self@ | "%(href)v",
                                .name * self@ | "%Di" trim,
                                .status * self@ | "%(title)v" sed "s/ (.*//",
                                .topics.u * self@ | "%(title)v" sed "s/.* (//; s/,.*//",
                                .posts.u * self@ | "%(title)v" sed "s/.*,//"
                            } | ,
                            .moderators * .moderators; a href; {
                                .user_link * self@ | "%(href)v",
                                .user [0] * self@ | "%Di" trim
                            } | ,
                            * .e>stats child@; { p child@, * self@ }; [0] * self@; {
                                .posts.u * self@ | "%i" sed 's/<.*//;s/,//g',
                                .topics.u * self@ | "%i" sed 's/.*>//;s/,//g'
                            },
                            .posts2.u [0] * .windowbg c@[0] child@ | "%i" sed 's/,//g',
                            .topics2.u [1] * .windowbg c@[0] child@ | "%i" sed 's/,//g',
                            * .lastpost child@; { p child@, * self@ }; [0] * self@; {
                                .lastpost * self@ -C@"[0] span .postby"; {
                                    [0] a href; {
                                        .user * self@ | "%Di" trim,
                                        .user_link * self@ | "%(href)v"
                                    },
                                    [1] a href; {
                                        .title * self@ | "%(title)Dv" / trim,
                                        .link * self@ | "%(href)v"
                                    },
                                    .date * self@ | "%t" tr "\n\t\r" trim sed "s/.*>//; s/([^ ]* ){3}//; s/^ *at //" "E"
                                },
                                .lastpost2 * self@ C@"[0] span .postby"; {
                                    [0] a href; {
                                        .title * self@ | "%(title)Dv" / trim,
                                        .link * self@ | "%(href)v"
                                    },
                                    [1] a href; {
                                        .user * self@ | "%Di" trim,
                                        .user_link * self@ | "%(href)v"
                                    },
                                    .date * self@ | "%t" trim sed "s/^at //"
                                }
                            }
                        } |
                    } | ,
                    .threads.u div #messageindex; {
                        table ( .table_grid )( .boardsframe ) child@; tbody child@; tr c@[4:] child@,
                        div #topic_container child@; div child@
                    }; {
                        .type1 * ( .icon1 )( .board_icon ) child@; img src | "%(src)v",
                        .type2 * .icon2 child@; img src | "%(src)v",
                        * ( .subject )( .info ) child@; {
                            [0] a href; {
                                .title * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            },
                            [1] a href; {
                                .user * self@ | "%Dt" trim,
                                .user_link * self@ | "%(href)v"
                            },
                            .lastpage.u [-] a ( .navPages )( .nav_page ) c@[0] m@v>"All" | "%i",
                            .icons.a {
                                * class self@,
                                div .icons; span .main_icons
                            }; * self@ | "%(class)v" / trim tr " " "\n" sed "/^subject$/d; /^windowbg/d; /^info$/d; /^info_block$/d; /^main_icons$/d; /^info_blockmain_icons$/d; /^lockmain_icons$/d",
                            .icons2.a img id child@ | "%(src)v\n" sed 's/.*\///; s/\..*//'
                        },
                        td .starter child@; a; {
                            .user2 [0] * self@ | "%Di" trim,
                            .user_link2 * self@ | "%(href)v"
                        },
                        {
                            td .stats child@,
                            * .board_stats child@; p
                        }; {
                            .replies.u * self@ | "%i" sed "s/<.*//;s/,//g;q",
                            .views.u * self@ | "%i" tr "\n\t\r," sed "s/.*>//"
                        },
                        .replies2.u td .replies c@[0] child@ | "%i" tr ",",
                        .views2.u td .views c@[0] child@ | "%i" tr ",",
                        * .lastpost child@; {
                            .lastpost [0] v>p child@; * rparent@; {
                                .link [0] a href | "%(href)v",
                                [1] a href; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                },
                                .date {
                                    span .smalltext child@,
                                    * self@
                                }; [0] * self@ | "%i" tr "\n\t\r" sed "s/<br.*//; s/.*<\/a>//; s/<[^>]*>//g; s/ by //" trim
                            },
                            .lastpost2 p child@; {
                                [0] a href; {
                                    .date * self@ | "%T" trim,
                                    .link * self@ | "%(href)v"
                                },
                                [1] a href; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                }
                            }
                        }
                    } |
                """
            )
        )

        categories = []

        prev_category_index = None
        prev_forums_index = None

        for i in t["categories"]:
            forums = i["forums"]

            if len(i["name"]) != 0:
                categories.append(i)
                prev_category_index = len(categories) - 1
                prev_forums_index = None
                i["forums"] = []

            if len(forums) == 0 or prev_category_index is None:
                continue

            out_forums = []

            for j in forums:
                for g in j["childboards"]:
                    g["link"] = url_merge(ref, g["link"])

                if len(j["link"]) == 0:
                    if prev_forums_index is not None and len(j["childboards"]) != 0:
                        out_forums[prev_forums_index]["childboards"] += j["childboards"]
                    continue

                j["link"] = url_merge(ref, j["link"])
                for g in j["moderators"]:
                    g["user_link"] = url_merge(ref, g["user_link"])

                redirects = 0

                lastpost = j["lastpost"]
                if len(lastpost["user_link"]) == 0 and len(lastpost["link"]) == 0:
                    lastpost = j["lastpost2"]
                j.pop("lastpost2")

                if len(lastpost["user_link"]) == 0 and len(lastpost["link"]) == 0:
                    redirects = j["posts"]
                    j["posts"] = 0
                    j["topics"] = 0
                j["redirects"] = redirects

                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
                j["lastpost"] = lastpost

                if j["posts"] == 0:
                    j["posts"] = j["posts2"]
                if j["topics"] == 0:
                    j["topics"] = j["topics"]
                j.pop("posts2")
                j.pop("topics2")

                out_forums.append(j)
                prev_forums_index = len(out_forums) - 1

            categories[prev_category_index]["forums"] += out_forums

        threads = t["threads"]

        for i in threads:
            if len(i["link"]) == 0:
                continue
            i["link"] = url_merge(ref, i["link"])
            i["type1"] = url_merge(ref, i["type1"])
            i["type2"] = url_merge(ref, i["type2"])

            if len(i["user_link"]) == 0:
                i["user"] = i["user2"]
                i["user_link"] = i["user_link2"]
            i.pop("user2")
            i.pop("user_link2")
            i["user_link"] = url_merge(ref, i["user_link"])

            i["icons"] += i["icons2"]
            i.pop("icons2")

            if i["replies"] == 0:
                i["replies"] = i["replies2"]
                i["views"] = i["views2"]
            i.pop("replies2")
            i.pop("views2")

            lastpost = i["lastpost"]
            if len(lastpost["link"]) == 0 and len(lastpost["user_link"]) == 0:
                lastpost = i["lastpost2"]
            i.pop("lastpost2")
            lastpost["link"] = url_merge(ref, lastpost["link"])
            lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
            i["lastpost"] = lastpost

        return {
            "format_version": "smf-2-forum",
            "url": url,
            "categories": categories,
            "threads": threads,
        }


class smf(ForumExtractorIdentify):
    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.trim = True

        self.v1 = smf1(self.session, **kwargs)
        self.v2 = smf2(self.session, **kwargs)

        self.extractors = [self.v1, self.v2]



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..enums import Outputs
from ..utils import dict_add, get_settings, url_merge_r, conv_short_size, url_merge
from ..exceptions import AlreadyVisitedError
from .common import ItemExtractor, ForumExtractor
from .identify import identify_invision


class invision(ForumExtractor):
    class User(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/(.*/)?(\d+)(-[^/]*)?/?"),
                    2,
                ),
            ]
            self.path_format = "m-{}"
            self.trim = True

        def get_url(self, url):
            url_delim = "?"
            if url.find("?") != -1:
                url_delim = "&"

            return "{}{}do=hovercard".format(url, url_delim)

        def get_first_html(
            self, url, settings, state, rq=None, ref=None, return_cookies=False
        ):
            settings = get_settings(
                settings, headers={"x-Requested-With": "XMLHttpRequest"}
            )
            return self.session.get_html(
                url, settings, state, self.trim, return_cookies
            )

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "invision-4-user", "url": url, "id": int(i_id)}

            t = json.loads(
                rq.search(
                    r"""
                .name h2 class=b>"ipsType_reset ipsType_"; a | "%Di" / trim,
                .background div .ipsCoverPhoto_container; img src | "%(src)v",
                .avatar img src .ipsUserPhoto | "%(src)v",
                .group p class="ipsType_reset ipsType_normal"; * c@[0] | "%i",

                .joined {
                    p class="ipsType_reset ipsType_medium ipsType_light" m@b>"Joined "; time datetime | "%(datetime)v",
                    div .cUserHovercard_data; li m@">Joined<"; time datetime | "%(datetime)v"
                },

                .lastseen {
                    p class="ipsType_reset ipsType_medium ipsType_light" m@b>"Last visited "; time datetime | "%(datetime)v",
                    div .cUserHovercard_data; li m@">Last visited<"; time datetime | "%(datetime)v"
                },

                .info dl; div l@[1]; {
                    .name dt | "%Di" / trim,
                    .value dd | "%i"
                } | ,
                div class=b>"ipsFlex ipsFlex-ai:center "; div class="ipsFlex ipsFlex-ai:center"; {
                    .rank img title | "%(title)v",
                    .rank_date time datetime | "%(datetime)v",
                },
                .badges.a div class=b>"ipsFlex ipsFlex-ai:center "; ul; li; img alt | "%(alt)v\n"
            """
                )
            )

            t["avatar"] = url_merge_r(ref, t["avatar"])
            t["background"] = url_merge_r(ref, t["background"])

            dict_add(ret, t)
            return ret

    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/(\d+)(-[^/]*)?/page/\d+/?"),
                    1,
                ),
                (
                    re.compile(r"/(.*/)?(\d+)(-[^/]*)?/?"),
                    2,
                ),
            ]
            self.trim = True

        def get_poll_answers(self, rq):
            ret = []
            for i in rq.filter(r"ul; li").self():
                el = {}
                el["option"] = i.search(r'div .ipsGrid_span4 | "%i"')
                el["votes"] = i.search(
                    r'div .ipsGrid_span1; * m@E>"^(<[^>]*>[^<]*</[^>]*>)?[^<]+$" | "%i" / sed "s/^<i.*<\/i> //"'
                )
                ret.append(el)
            return ret

        def get_poll_questions(self, rq):
            ret = []
            for i in rq.filter(r"ol .ipsList_reset .cPollList; li l@[1]").self():
                el = {}
                el["question"] = i.search(r'h3; span | "%i"')
                el["answers"] = self.get_poll_answers(i)
                ret.append(el)

            return ret

        def get_poll(self, rq):
            ret = {}
            controller = rq.filter(r"section data-controller=core.front.core.poll")

            title = ""
            questions = []

            if controller:
                title = controller.search(
                    r'h2 l@[1]; span l@[1] | "%i" / sed "s/<.*//;s/&nbsp;//g;s/ *$//"'
                )
                questions = self.get_poll_questions(controller)

            ret["title"] = title
            ret["questions"] = questions
            return ret

        def get_reactions_details(self, rq, settings, state):
            ret = []
            nexturl = rq.search(
                r'ul .ipsReact_reactions; li .ipsReact_reactCount; [0] a href | "%(href)v" / sed "s/&amp;/\&/g; s/&reaction=.*$//;q" "E"'
            )

            nsettings = get_settings(
                settings, headers={"x-Requested-With": "XMLHttpRequest"}
            )

            while True:
                if len(nexturl) == 0:
                    break

                rq, ref = self.session.get_html(
                    nexturl,
                    nsettings,
                    state,
                    True,
                )

                t = json.loads(
                    rq.search(
                        r"""
                        .reactions ol; li; {
                            .avatar a .ipsUserPhoto; img src | "%(src)v",
                            a .ipsType_break href; {
                                .user_link * l@[0] | "%(href)v",
                                .user * l@[0] | "%Di" trim
                            },
                            .reaction span .ipsType_light; img src | "%(src)v" / sed "s#.*/reactions/##;s/\..*//;s/^react_//",
                            .date time datetime | "%(datetime)v"
                        } |
                        """
                    )
                )

                if len(ret) == 0:
                    self.state_add_url("reactions", nexturl, state, settings)

                for i in t["reactions"]:
                    i["avatar"] = url_merge_r(nexturl, i["avatar"])
                    i["user_link"] = url_merge_r(nexturl, i["user_link"])

                ret += t["reactions"]

                nexturl = rq.search(
                    r'li .ipsPagination_next -.ipsPagination_inactive; [0] a href | "%(href)v" / sed "s/&amp;/&/; s/&reaction=.*$//;q"'
                )

            return ret

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "invision-4-thread", "url": url, "id": int(i_id)}
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                div #ipsLayout_mainArea; h1 class="ipsType_pageTitle ipsContained_container"; {
                    .title span class="ipsType_break ipsContained"; span -class | "%Di" trim,
                    .badges.a span title | "%(title)v\n"
                },
                .rating.b ul .ipsRating_collective; li .ipsRating_on | "true",
                div .ipsPageHeader; {
                    .user_link a .ipsType_break href | "%(href)v",
                    .user a .ipsType_break href; * c@[0] | "%Di" trim,
                    .user_avatar a .ipsUserPhoto; img src | "%(src)v",
                    .user_followers.u a .ipsFollow; span .ipsCommentCount | "%i" / tr ",.",
                },
                .date div .ipsFlex-flex:11; time datetime | "%(datetime)v",
                .path.a nav class=b>"ipsBreadcrumb ipsBreadcrumb_top "; ul data-role="breadcrumbList"; li; a; span | "%Dt\n" / sed "s/ $//" trim,

                .tags.a {
                    div .ipsPageHeader; ul .ipsTags; a; span c@[0] | "%i",
                    div #ipsLayout_mainArea; h1 class="ipsType_pageTitle ipsContained_container"; a .ipsTag_prefix rel=tag; span c@[0] | "%i",
                },

                .warning div .cTopicPostArea; span .ipsType_warning | "%i",
            """
                )
            )
            dict_add(ret, t)

            ret["poll"] = self.get_poll(rq)
            ret["user_link"] = url_merge_r(ref, ret["user_link"])
            ret["user_avatar"] = url_merge_r(ref, ret["user_avatar"])

            t = json.loads(
                rq.search(
                    r"""
                .recommended div data-role=recommendedComments; div .ipsBox data-commentID; {
                    .id.u div .ipsBox data-commentID | "%(data-commentID)v",
                     div .ipsReactOverview; {
                        .reactions.a ul; li; img alt | "%(alt)v\n",
                        .reaction_count.u p class="ipsType_reset ipsType_center" | "%i"
                    },
                    .user_avatar aside; img src | "%(src)v",
                    div .ipsColumn; {
                        * .ipsComment_meta; {
                            .user_link a .ipsType_break href | "%(href)v",
                            .user a .ipsType_break href; * c@[0] | "%Di" trim,
                            .date time datetime | "%(datetime)v"
                        },
                        .link a .ipsButton href | "%(href)v",
                        a .ipsType_break href c@[0] l@[1]; {
                            .ruser_link a l@[0] | "%(href)v",
                            .ruser a l@[0] | "%i"
                        },
                        .content div .ipsType_richText | "%i"
                    }
                }
            """
                )
            )
            rec = t["recommended"]
            rec["user_avatar"] = url_merge_r(ref, rec["user_avatar"])
            rec["user_link"] = url_merge_r(ref, rec["user_link"])
            rec["link"] = url_merge_r(ref, rec["link"])
            rec["ruser_link"] = url_merge_r(ref, rec["ruser_link"])
            dict_add(ret, t)

            expr = reliq.expr(
                r"""
                .id.u article #B>elComment_[0-9]* | "%(id)v\n" / sed "s/^elComment_//",
                aside; {
                    .user h3 class=b>"ipsType_sectionHead cAuthorPane_author "; * c@[0] [0] | "%Di" trim,
                    * .cAuthorPane_photo data-role=photo; {
                        .user_avatar a .ipsUserPhoto; img src | "%(src)v",
                        .badges.a {
                            img title alt | "%(alt)v\n",
                            span .cAuthorPane_badge title=e>" joined recently" | "%(title)v\n"
                        }
                    },
                    ul .cAuthorPane_info; {
                        .group li data-role=group; * c@[0] | "%i",
                        .group_icon li data-role=group-icon; img src | "%(src)v",
                        .rank_title li data-role=rank-title | "%i",
                        .rank_image li data-role=rank-image; * | "%i",
                        .reputation_badge li data-role=reputation-badge; span | "%i" / sed "s/^<i .*<\/i> //;s/,//g;q",
                        .posts li data-role=posts | "%i" / sed "s/ .*//;s/,//g",
                        .custom li data-role=custom-field | "%i",
                        .user_info.a ul .ipsList_reset; li; a title l@[1] | "%(title)v\n" / sort "u",
                    }
                },
                div .ipsComment_meta; {
                    .top_badges.a div class=a>"ipsComment_badges"; ul .ipsList_reset; li; strong | "%i\n" / sed "s#<i [^>]*></i> ##g",
                    .date time datetime | "%(datetime)v",
                },
                .content div .cPost_contentWrap; div data-role=commentContent | "%i",
                .signature div data-role=memberSignature; div data-ipslazyload | "%i",
            """
            )
            posts = []

            while True:
                for i in rq.filter(r"article #B>elComment_[0-9]*").self():
                    post = {}

                    post["user_link"] = url_merge_r(
                        ref,
                        i.search(
                            r'aside; h3 class=b>"ipsType_sectionHead cAuthorPane_author "; a href | "%(href)v"'
                        ),
                    )

                    user_link = post["user_link"]
                    if Outputs.users in settings["output"] and len(user_link) > 0:
                        try:
                            self.user.get("users", user_link, settings, state)
                        except self.common_exceptions as ex:
                            self.handle_error(ex, user_link, settings, True)

                    t = json.loads(i.search(expr))
                    t["user_avatar"] = url_merge_r(ref, t["user_avatar"])
                    t["group_icon"] = url_merge_r(ref, t["group_icon"])
                    t["rank_image"] = url_merge_r(ref, t["rank_image"])
                    dict_add(post, t)

                    t = json.loads(
                        i.search(
                            r"""
                        ul .ipsReact_reactions; {
                            .reactions_users li .ipsReact_overview; a href -href=a>?do= l@[1]; {
                                .link * l@[0] | "%(href)v",
                                .name * l@[0] | "%i"
                            } | ,
                            .reactions_temp.a li .ipsReact_reactCount; span a@[0] / sed "s/<span><img .* alt=\"//; s/\".*//; N; s/\n/\t/; s/<span>//g; s/<\/span>//g"
                        }
                    """
                        )
                    )
                    for j in t["reactions_users"]:
                        j["link"] = url_merge_r(ref, j["link"])
                    t["reactions"] = []
                    for j in t["reactions_temp"]:
                        el = {}
                        reaction = j.split("\t")
                        el["name"] = reaction[0]
                        el["count"] = 1
                        try:
                            el["count"] = int(reaction[1])
                        except Exception:
                            pass
                        t["reactions"].append(el)
                    t.pop("reactions_temp")
                    dict_add(post, t)

                    reactions_details = []
                    if Outputs.reactions in settings["output"]:
                        try:
                            reactions_details = self.get_reactions_details(
                                i, settings, state
                            )
                        except self.common_exceptions as ex:
                            self.handle_error(ex, i, settings, True)
                    post["reactions_details"] = reactions_details

                    posts.append(post)

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break

                try:
                    rq, ref = self.session.get_html(nexturl, settings, state, True)
                except AlreadyVisitedError:
                    break

            ret["posts"] = posts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_invision

        self.trim = True

        self.thread = self.Thread(self.session)
        self.user = self.User(self.session)
        self.thread.user = self.user
        self.thread.get_next = self.get_next

        self.board_forums_expr = reliq.expr(
            r'li class=b>"cForumRow ipsDataItem "; div .ipsDataItem_main; h4; a href | "%(href)v\n", div .ipsForumGrid; a .cForumGrid__hero-link href | "%(href)v\n"'
        )
        self.forum_forums_expr = self.board_forums_expr
        self.forum_threads_expr = reliq.expr(
            r'ol data-role=tableRows; h4; a class="" href=e>"/" | "%(href)v\n"'
        )
        self.guesslist = [
            {
                "func": "get_thread",
                "exprs": [r"/(.*[/?])?(thread|topic)s?/"],
            },
            {
                "func": "get_forum",
                "exprs": [r"/(.*[/?])?forums?/"],
            },
            {"func": "get_board", "exprs": None},
        ]

        self.findroot_expr = reliq.expr(
            r"""
            {
                [0] ul data-role=breadcrumbList; a href | "%(href)v\n" / sed "
                    1!G
                    h
                    \#/(((forum|foro|board)s?|community)/?|index\.php)$#{p;q}
                   $p" "En" line [-],
               li #b>elNavSecondary_ i>data-navapp=i>forums; [0] a href | "%(href)v",
               li #b>lmgNavSub_; [0] a href m@i>"forums" | "%(href)v\n",
               nav; li .ipsMenu_item; [0] a m@i>forums href | "%(href)v\n",
               [0] a href=Ea>"(/|^)((forum|foro)s?|community|communaute|comunidad|ipb)(\.([a-zA-Z0-9.-]\.)*[a-zA-Z]/?|/?$)" | "%(href)v\n",
               [0] a href=Ea>"(/|^)(index|index\.php)/?$" | "%(href)v\n"
            } / line [0] tr "\n"
           """
        )
        self.findroot_board = True
        self.findroot_board_expr = re.compile(
            r"^(/[^\.-])?/((forum|foro|board)s?|index\.php|community|communaute|comunidad|ipb)/?$"
        )

    def get_forum_next_page(self, rq):
        return rq.search(
            'ul .ipsPagination; li .ipsPagination_next -.ipsPagination_inactive; [0] a | "%(href)v"'
        )

    def get_next_page(self, rq):
        return rq.search(
            r'ul .ipsPagination [0]; li .ipsPagination_next -.ipsPagination_inactive; a | "%(href)v" / sed "s#/page/([0-9]+)/.*#/?page=\1#" "E"'
        )

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                .categories [0] * .cForumList; {
                    li child@,
                    * -C@"[0] li l@[1]" self@
                }; {
                    h2 child@; {
                        .name * -title c@[0] m@>[1:] | "%Di" / trim,
                        .link [-] a | "%(href)v"
                    },
                    .forums {
                        ol .ipsDataList child@; li child@,
                        div .ipsForumGrid; div child@
                    }; {
                        [0] div ( .ipsDataItem_icon )( .cForumGrid__icon ); {
                            .status [0] span class | "%(class)v" sed "s/.*cForumIcon_//; s/ .*//",
                            .icon [0] img src | "%(src)v"
                        },
                        .icon2 [0] span .cForumGrid__hero-image data-background-src | "%(data-background-src)v",
                        div ( .ipsDataItem_main )( .cForumGrid__content ); {
                            * ( .ipsDataItem_title )( .cForumGrid__title ); [0] a; {
                                .title * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            },
                            .description [0] * .ipsType_richText | "%T" trim,
                        },
                        .posts {
                            [0] dt .ipsDataItem_stats_number | "%i",
                            [0] ul .cForumGrid__title-stats; [0] li c@[0] | "%i" sed "s/ .*//"
                        },
                        .followers [0] ul .cForumGrid__title-stats; [0] a c@[0] | "%i" sed "s/ .*//",
                        .childboards ul ( .ipsDataItem_subList )( .cForumGrid__subforums ); a; {
                            .name * self@ | "%Di" / trim,
                            .link * self@ | "%(href)v"
                        } | ,
                        .lastpost [0] * ( .ipsDataItem_lastPoster )( .cForumGrid__last ); {
                            .avatar * .ipsUserPhoto; [0] img src | "%(src)v",
                            * .ipsDataItem_lastPoster__title; [0] a; {
                                .title * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            },
                            li .ipsType_light; [0] a .ipsType_break; {
                                .user * c@[0] | "%Di" trim,
                                .user_link * self@ | "%(href)v"
                            },
                            .date time datetime | "%(datetime)v"
                        }
                    } |
                } | ,
                .threads [0] ol .cTopicList; li child@; {
                    .avatar div .ipsTopicSnippet__avatar; [0] img | "%(src)v",
                    [0] * .ipsDataItem_title; {
                        .icons.a i | "%(class)v\n" / sed "s/.*fa-//",
                        [0] a -rel=tag; {
                            .title * c@[0] | "%Di" / trim,
                            .link * self@ | "%(href)v",
                        }
                    },
                    [0] div ( .ipsDataItem_meta )( .ipsTopicSnippet__date ); {
                        .date time datetime | "%(datetime)v",
                        [0] a .ipsType_break; {
                            .user * c@[0] | "%Di" trim,
                            .user_link * self@ | "%(href)v"
                        }
                    },
                    .lastpage.u * .ipsPagination; [-] a | "%i",
                    .tags a rel=tag; {
                        .name * c@[0] | "%Di" / trim,
                        .link * self@ | "%(href)v"
                    } | ,
                    [0] * ( .ipsDataItem_stats )( .ipsTopicSnippet__stats ); {
                        .group-indicator [0] img .cGroupIndicator src | "%(src)v",
                        span .ipsDataItem_stats_number; {
                            .replies [0] * self@ | "%i",
                            .views [1] * self@ | "%i",
                        }
                    },
                    .snippet div .ipsTopicSnippet__snippet; * c@[0] | "%i",
                    .lastpost [0] * ( .ipsDataItem_lastPoster )( .ipsTopicSnippet__last ); {
                        .avatar [0] * .ipsUserPhoto; [0] img src | "%(src)v",
                        [0] a .ipsType_break; {
                            .user * c@[0] | "%Di" trim,
                            .user_link * self@ | "%(href)v"
                        },
                        .date time datetime | "%(datetime)v"
                    }
                } |
                """
            )
        )

        categories = t["categories"]

        for i in categories:
            i["link"] = url_merge(ref, i["link"])

            for j in i["forums"]:
                for g in j["childboards"]:
                    g["link"] = url_merge(ref, g["link"])

                j["link"] = url_merge(ref, j["link"])

                if len(j["icon"]) == 0:
                    j["icon"] = j["icon2"]
                j.pop("icon2")
                j["icon"] = url_merge(ref, j["icon"])

                j["posts"] = conv_short_size(j["posts"])
                j["followers"] = conv_short_size(j["followers"])

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
                lastpost["avatar"] = url_merge(ref, lastpost["avatar"])

        threads = t["threads"]

        for i in threads:
            i["link"] = url_merge(ref, i["link"])
            i["user_link"] = url_merge(ref, i["user_link"])
            i["avatar"] = url_merge(ref, i["avatar"])
            i["group-indicator"] = url_merge(ref, i["group-indicator"])

            i["replies"] = conv_short_size(i["replies"])
            i["views"] = conv_short_size(i["views"])

            for j in i["tags"]:
                j["link"] == url_merge(ref, j["link"])

            lastpost = i["lastpost"]
            lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
            lastpost["avatar"] = url_merge(ref, lastpost["avatar"])

        return {
            "format_version": "invision-4-forum",
            "url": url,
            "categories": categories,
            "threads": threads,
        }



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

from .common import ForumExtractorIdentify

from .smf import smf, smf1, smf2
from .phpbb import phpbb
from .xmb import xmb
from .xenforo import xenforo, xenforo1, xenforo2
from .invision import invision
from .hackernews import hackernews
from .stackexchange import stackexchange
from .vbulletin import vbulletin


class Extractor(ForumExtractorIdentify):
    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.trim = True

        self.smf = smf(self.session, **self.settings)
        self.phpbb = phpbb(self.session, **self.settings)
        self.xmb = xmb(self.session, **self.settings)
        self.xenforo = xenforo(self.session, **self.settings)
        self.invision = invision(self.session, **self.settings)
        self.hackernews = hackernews(self.session, **self.settings)
        self.stackexchange = stackexchange(self.session, **self.settings)
        self.vbulletin = vbulletin(self.session, **self.settings)

        self.extractors = [
            self.hackernews,
            self.phpbb,
            self.invision,
            self.xmb,
            self.xenforo,
            self.smf,
            self.stackexchange,
            self.vbulletin,
        ]



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import re
import json
from reliq import reliq

from ..enums import Outputs
from ..utils import dict_add, url_merge, url_merge_r, conv_short_size
from .common import ItemExtractor, ForumExtractor, ForumExtractorIdentify
from .identify import identify_xenforo1, identify_xenforo2


guesslist = [
    {
        "func": "get_thread",
        "exprs": [r"/(.*[/?])?(thread|topic|tema)s?/", r"/t/."],
    },
    {"func": "get_board", "exprs": [r"^/(.*[/?])?forums?/$"]},
    {
        "func": "get_forum",
        "exprs": [r"/(.*[/?])?forums?/", r"/f/."],
    },
    {
        "func": "get_tag",
        "exprs": [r"/(.*[/?])?tags?/"],
    },
    {"func": "get_board", "exprs": None},
]


class xenforo2(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/(.*[./])?(\d+)(/.*)?$"),
                    2,
                )
            ]

        def get_search_user(self, rq, state, ref, first_delim, xfToken, settings):
            user_url = rq.search(
                r"""
                {
                    h4 class=b>message-name,
                    div .MessageCard__avatar
                }; [0] a data-user-id href | "%(href)v"
            """
            )
            if len(user_url) > 0:
                self.user.get(
                    "users",
                    "{}{}tooltip=true&_xfWithData=1&_xfToken={}&_xfResponseType=json".format(
                        url_merge(ref, user_url), first_delim, xfToken
                    ),
                    settings,
                    state,
                )

        def get_xfToken(self, rq):
            xfToken = rq.search(r'html data-csrf | "%(data-csrf)v"')
            if len(xfToken) == 0:
                xfToken = rq.search(r'[0] input name="_xfToken" value | "%(value)v"')
            if len(xfToken) == 0:
                xfToken = self.session.headers.get("csrf")

            return xfToken

        def get_reactions(self, rq, ref, first_delim, xfToken, settings, state):
            ret = []
            reactions_url = rq.search(
                r'{ a .reactionsBar-link href | "%(href)v\n", div #b>reactions-bar-; a .list-reacts href | "%(href)v\n" } / line [0] tr "\n"'
            )

            if len(reactions_url) > 0:
                reactions_url = "{}{}_xfRequestUri=&_xfWithData=1&_xfToken={}&_xfResponseType=json".format(
                    url_merge(ref, reactions_url), first_delim, xfToken
                )

                obj = reliq(
                    self.session.get_json(reactions_url, settings, state)["html"][
                        "content"
                    ].translate(str.maketrans("", "", "\n\t"))
                )

                t = json.loads(
                    obj.search(
                        r"""
                    .reactions div .contentRow; {
                        .user_id.u * .username data-user-id | "%(data-user-id)v",
                        .user * .username data-user-id; * c@[0] | "%Di" / trim,
                        .date time .u-dt datetime | "%(datetime)v",
                        .reaction span .reaction; img title | "%(title)v"
                    } |
                """
                    )
                )
                ret = t["reactions"]

                self.state_add_url("reactions", reactions_url, state, settings)

            return ret

        def get_contents(self, rq, settings, state, url, ref, i_id):
            page = 0
            url_first_delimiter = "?"
            if url.find("?") != -1:
                url_first_delimiter = "&"
            ret = {"format_version": "xenforo-2-thread", "url": url, "id": int(i_id)}

            t = json.loads(
                rq.search(
                    r"""
                .title h1; {
                    h1 .p-title-value | "%Di",
                    h1 qid="page-header" | "%Di",
                    h1 .MessageCard__thread-title | "%Di"
                } / sed ":x; s/<[^>]*>//g; $!{N;s/\n/ /;bx}" trim,

                div .p-description; {
                    .user_id.u * data-user-id | "%(data-user-id)v",
                    .user * data-user-id; * c@[0] | "%Di" / trim,
                    .date [0] time datetime | "%(datetime)v",
                },
                .path.a [0] ul .p-breadcrumbs -.p-breadcrumbs--bottom; span | "%Di\n" / trim "\n",
                .tags.a("|") a class=b>tagItem | "%i\n" / sed ":x; s/\t//g; /^$/d; $!{N;s/\n/|/;bx}; s/|$//; s/|\+/|/g" tr "\n",
                .poll form data-xf-init="poll-block ajax-submit"; {
                    .title h2 .block-header | "%Di" / sed "s/\t//g; s/<[^>]*>//g; s/^ *//; s/ *$//; /^$/d;" trim,
                    .answers [:-1] li; {
                        .option h3 .pollResult-response | "%i",
                        .votes.u span .pollResult-votes | "%i" / sed "s/\t//g; s/<[^>]*>//g; s/^ *//; s/ *$//; /^$/d; s/^.* //"
                    } |
                }
            """
                )
            )
            dict_add(ret, t)

            xfToken = self.get_xfToken(rq)

            posts = []
            expr = reliq.expr(
                r"""
                {
                    h4 class=b>message-name,
                    * .MessageCard__user-info__name
                }; {
                    .user [0] * c@[0] | "%Di" trim,
                    .user_link [0] a class href | "%(href)v",
                    .user_id.u [0] * data-user-id | "%(data-user-id)v"
                },

                .user_avatar {
                    div .message-avatar,
                    div .MessageCard__avatar
                }; img src | "%(src)v",

                .user_title h5 .userTitle | "%i",

                .user_banners.a div .userBanner; strong | "%i\n",

                .user_extras {
                    .pairs1 div .message-userExtras; dl l@[1]; {
                        .key dt; {
                            * title | "%(title)v\a",
                            * l@[0] | "%i"
                        } / sed "s/^\a*//;s/\a.*//",
                        .value dd | "%i",
                    } | ,

                    .pairs2 div .message-userExtras; div .pairs; span title; {
                        .key * l@[0] | "%(title)v",
                        .value * l@[0] | "%i",
                    } | ,

                    .pairs3 div .MessageCard__user-details; span class -.MessageCard__dot-separator; {
                        .key * l@[0] | "%(class)v" / sed 's/^MessageCard__//',
                        .value * l@[0] | "%i"
                    } | ,

                    div .message-userExtras; {
                        .stars.a ul .reputation-star-container; li .eB>"[a-z]star" | "%(class)v\n" / sed "s/.* //",

                        .bar.a ul .reputation-bar-container; li .reputation-bar | "%(class)v\n" / sed "s/.* //"
                    }
                },

                .id.u [0] * #E>(js-)?post-[0-9]+ | "%(id)v" / sed "s/^js-//; s/^post-//;q",

                .date {
                    * class=b>message-attribution-main; time datetime | "%(datetime)v\t",
                    * .MessageCard__date-created; [0] time datetime .u-dt | "%(datetime)v"
                } / sed "s/\t.*//",

                .text {
                    article class=b>message-body,
                    div .MessageCard__content-inner
                }; div .bbWrapper | "%i",

                .attachments.a ul .attachmentList; a,

                .signature div #signature-content-wrapper; div .bbWrapper | "%i"
            """
            )

            rq = reliq(rq.get_data().translate(str.maketrans("", "", "\n\t\r\a")))

            while True:
                post_tags = rq.search(
                    r"""
                        [0] div c@[2:] .california-article-post,
                        div #E>(js-)?post-[0-9]+ ||
                        div #thread-main-section; div .MessageCard l@[1] | "%i\n" ||
                        article id ||
                        div .block-container; div .MessageCard | "%i\n"
                """
                ).split("\n")[:-1]

                for i in post_tags:
                    tag = reliq(i)

                    post = json.loads(tag.search(expr))

                    post["user_link"] = url_merge_r(ref, post["user_link"])
                    if post["user_avatar"][:1] == "/":
                        post["user_avatar"] = re.sub(
                            r"\?[0-9]+$", r"", post["user_avatar"]
                        )
                    post["user_avatar"] = url_merge_r(ref, post["user_avatar"])

                    uext = post["user_extras"]
                    uext["pairs"] = uext["pairs1"] + uext["pairs2"] + uext["pairs3"]
                    uext.pop("pairs1")
                    uext.pop("pairs2")
                    uext.pop("pairs3")

                    reactions = []

                    if len(xfToken) > 0:
                        try:
                            if Outputs.users in settings["output"]:
                                self.get_search_user(
                                    tag,
                                    state,
                                    ref,
                                    url_first_delimiter,
                                    xfToken,
                                    settings,
                                )
                        except self.common_exceptions as ex:
                            self.handle_error(
                                ex,
                                "{}{}{}".format(url, url_first_delimiter, xfToken),
                                settings,
                                True,
                            )

                        try:
                            if Outputs.reactions in settings["output"]:
                                reactions = self.get_reactions(
                                    tag,
                                    ref,
                                    url_first_delimiter,
                                    xfToken,
                                    settings,
                                    state,
                                )
                        except self.common_exceptions as ex:
                            self.handle_error(
                                ex,
                                "{}{}{}".format(url, url_first_delimiter, xfToken),
                                settings,
                                True,
                            )

                    post["reactions"] = reactions

                    posts.append(post)

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state, True)

            ret["posts"] = posts
            return ret

    class User(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/(.*[./])?(\d+)/[\&?]tooltip=true\&"),
                    2,
                )
            ]
            self.path_format = "m-{}"

        def get_first_html(self, url, settings, state, rq=None, ref=None):
            rq = reliq(self.session.get_json(url, settings, state)["html"]["content"])
            if ref is None:
                ref = self.session.base(rq, url)
            return (rq, ref)

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "xenforo-2-user", "url": url, "id": int(i_id)}

            t = json.loads(
                rq.search(
                    r"""
                .background div class=B>"memberProfileBanner memberTooltip-header.*" style=a>"url(" | "%(style)v" / sed "s#.*url(##;s#^//#https://#;s/?.*//;p;q" "n",
                .location a href=b>/misc/location-info | "%i",
                .avatar img src | "%(src)v" / sed "s/?.*//; q",
                .title span .userTitle | "%Di" / trim,
                .banners.a * .userBanner; strong | "%i\n",
                .name h4 .memberTooltip-name; a; * c@[0] | "%Di" / trim,
                .forum em; a href | "%(href)v",
                .extras dl .pairs c@[!0]; {
                    .key dt | "%i",
                    .value dd; {
                        time datetime | "%(datetime)v\a",
                        a m@v>"<" | "%i\a",
                        * l@[0] | "%i"
                    } / tr '\n' sed "s/^\a*//;s/\a.*//"
                } |
            """
                )
            )
            t["background"] = url_merge_r(ref, t["background"])
            t["avatar"] = url_merge_r(ref, t["avatar"])
            dict_add(ret, t)

            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_xenforo2

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next
        self.user = self.User(self.session)
        self.thread.user = self.user

        self.board_forums_expr = reliq.expr(
            r'* .node-title; a href -href=b>/link-forums/ | "%(href)v\n"'
        )
        self.forum_forums_expr = self.board_forums_expr
        self.forum_threads_expr = reliq.expr(
            r'* .structItem-title; a -.labelLink href | "%(href)v\n"'
        )
        self.tag_threads_expr = reliq.expr(
            r'div class=b>contentRow; a -.labelLink href -data-user-id | "%(href)v\n"'
        )
        self.guesslist = guesslist

        self.findroot_expr = reliq.expr(
            r"""
            {
                * .p-nav; [0] a href | "%(href)v\n",
                * #header-forum-listing href | "%(href)v\n"
            } / line [0] tr "\n"
            """
        )
        self.findroot_board = True
        self.findroot_board_expr = re.compile(
            r"^(/[^\.-])?/((forum|foro|board)s?|index\.php|community|communaute|comunidad)/?$",
        )

    def get_next_page(self, rq):
        url = rq.search(
            r'div .block-outer; [0] a .pageNav-jump .pageNav-jump--next href | "%(href)v"'
        )
        if not re.search(r"/page-[0-9]*(\?.*)?/?(#.*)?$", url):
            return ""
        return url

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_tag_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                .categories div .block -b>data-widget- -.thNodes__nodeList -data-type=thread; {
                    [0] * ( .block-header )( .section-header ); [0] * c@[0]; {
                        .name * self@ | "%Di" / trim,
                        .link * self@ | "%(href)v"
                    },
                    .forums [0] div ( .block-body )( .node-list ); div .node l@[1:2]; {
                        .icon [0] span .node-icon; {
                            i | "%(class)v" sed "s/.* //",
                            span .icon | "%(class)v" tr " " "\n" sed "/^icon$/d;/^forum$/d;s/node--//" trim,
                        },
                        * .node-main; {
                            * .node-title; [0] a; {
                                .name * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            },
                            .description div .node-description | "%i",
                            .childboards [0] ol ( .node-subNodeFlatList )( .subNodeMenu ); a .subNodeLink; {
                                .icon {
                                    * self@ | "%(class)v\a" sed "s/.*--//; /^subNodeLink/d",
                                    i | "%(class)v" sed "s/ subNodeLink-icon//; s/.* //"
                                } / sed "s/\a.*//",
                                .name * self@ | "%Dt" trim,
                                .link * self@ | "%(href)v"
                            } |
                        },
                        div ( .node-meta )( .node-stats ); {
                            dd; {
                                .topics [0] * self@ | "%i",
                                .posts [1] * self@ | "%i"
                            },
                            span title; {
                                .posts2.u [0] * self@ | "%(title)v" tr ",. ",
                                .views.u [1] * self@ | "%(title)v" tr ",. "
                            }
                        },
                        .placeholder [0] * .node-extra-placeholder | "%i",
                        .lastpost div .node-extra; {
                            .avatar div .node-extra-icon; [0] img | "%(src)v",
                            * .node-extra-title; [0] a; {
                                .title * self@ | "%(title)Dv" trim,
                                .link * self@ | "%(href)v",
                                .label [0] span .label | "%i"
                            },
                            .date time .node-extra-date datetime | "%(datetime)v",
                            * .node-extra-user; [0] a; {
                                .user [0] * c@[0] | "%Di" trim,
                                .user_link * self@ | "%(href)v"
                            }
                        },
                        .date2 [0] * .last-time; time datetime | "%(datetime)v"
                    } |
                } | ,
                .threads div .structItem--thread; {
                    .votes.u span .contentVote-score | "%i",
                    .avatar div .structItem-iconContainer; [0] img src | "%(src)v",
                    div .structItem-cell--main; {
                        .icons.a {
                            ul .structItem-statuses; span; {
                                img title | "%(title)v\n",
                                * self@ | "%t\n" sed "/^$/d",
                            },
                            svg title | "%(title)v\n"
                        },
                        .label span .label | "%t",
                        * .structItem-title; [-] a -.labelLink; {
                            .title * self@ | "%Di" trim,
                            .link * self@ | "%(href)v"
                        },
                        [0] a .username; {
                            .user * c@[0] | "%Di" trim,
                            .user_link * self@ | "%(href)v"
                        },
                        .date [0] time datetime | "%(datetime)v",
                        .lastpage.u * .structItem-pageJump; [-] a | "%i"
                    },
                    div .structItem-cell--meta; {
                        dt m@B>"^[0-9]",
                        dd
                    }; {
                        .replies [0] * c@[0] | "%i",
                        .views [1] * c@[0] | "%i"
                    },
                    .replies2.u [0] div .reply-count title | "%(title)v" tr ",. ",
                    .views2.u [0] div .view-count title | "%(title)v" tr ",. ",
                    .lastpost div ( .structItem-cell--latest )( .last-post-cell ); {
                        .date [0] time datetime | "%(datetime)v",
                        [0] a .username; {
                            .user * c@[0] | "%Di" trim,
                            .user_link * self@ | "%(href)v"
                        },
                    },
                    .lp-avatar div .structItem-cell--iconEnd; [0] img src | "%(src)v"
                } |
                """
            )
        )

        categories = []

        for i in t["categories"]:
            if len(i["name"]) == 0 and len(i["forums"]) == 0:
                continue

            i["link"] = url_merge(ref, i["link"])

            for j in i["forums"]:
                for g in j["childboards"]:
                    g["link"] = url_merge(ref, g["link"])

                j["link"] = url_merge(ref, j["link"])

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
                lastpost["avatar"] = url_merge(ref, lastpost["avatar"])
                if len(lastpost["date"]) == 0:
                    lastpost["date"] = j["date2"]
                j.pop("date2")

                if len(j["posts"]) == 0:
                    j["topics"] = 0
                    j["posts"] = j["posts2"]
                else:
                    j["topics"] = conv_short_size(j["topics"])
                    j["posts"] = conv_short_size(j["posts"])
                j.pop("posts2")

            categories.append(i)

        threads = t["threads"]

        for i in threads:
            i["link"] = url_merge(ref, i["link"])
            i["avatar"] = url_merge(ref, i["avatar"])
            i["user_link"] = url_merge(ref, i["user_link"])

            if len(i["replies"]) == 0:
                i["replies"] = i["replies2"]
                i["views"] = i["views2"]
            else:
                i["replies"] = conv_short_size(i["replies"])
                i["views"] = conv_short_size(i["views"])
            i.pop("replies2")
            i.pop("views2")

            lastpost = i["lastpost"]
            lastpost["user_link"] = url_merge(ref, lastpost["user_link"])
            lastpost["avatar"] = url_merge(ref, i["lp-avatar"])
            i.pop("lp-avatar")

        return {
            "format_version": "xenforo-2-forum",
            "url": url,
            "categories": categories,
            "threads": threads,
        }


class xenforo1(ForumExtractor):
    class Thread(ItemExtractor):
        def __init__(self, session):
            super().__init__(session)

            self.match = [
                (
                    re.compile(r"/(.*[./])?t?(\d+)(/(\?.*)?|\.html)?$"),
                    2,
                )
            ]

        def get_avatar_and_userid(self, ref, messageUB):
            user_id = "0"
            avatar = messageUB.search(r'* class=b>avatar; [0] img src | "%(src)v"')
            if len(avatar) == 0:
                user_id = messageUB.search(
                    r'span class="img m" style | "%(style)v" / sed "s#^.*/avatars/m/[0-9]*/##; s#\..*##"'
                )
                avatar = messageUB.search(
                    r"""span class="img m" style | "%(style)v" / sed "s/?[0-9]*')//; s/background-image: url('/\//; s/')$//" """
                )
            else:
                r = re.search(r"/(\d+)\.[a-zA-Z]*(\?.*)?$", avatar)
                if r:
                    user_id = r[1]

            avatar = re.sub(r"\?[0-9]+$", r"", avatar)
            avatar = url_merge_r(ref, avatar)

            if user_id == "0":
                user_id = messageUB.search(
                    r'[0] * .userText; a .username href | "%(href)v" / sed "s#/$##;s/.*\.//"'
                )
            try:
                user_id = int(user_id)
            except ValueError:
                user_id = 0

            return avatar, user_id

        def get_contents(self, rq, settings, state, url, ref, i_id):
            ret = {"format_version": "xenforo-1-thread", "url": url, "id": int(i_id)}
            page = 0

            t = json.loads(
                rq.search(
                    r"""
                .title { div class=b>titleBar; [0] h1 | "%Di" trim, div #header; [0] h1 | "%Di" trim / sed ":x; s/<[^>]*>//g; $!{N;s/\n/ /;bx}" },
                p #pageDescription; {
                    .user_id.u a .username href | "%(href)v" / sed "s/^.*[\/.]\([0-9]\+\)/\1/; s/[^0-9]$//",
                    .user [0] a .username | "%Di" trim,
                    .date [0] * .DateTime | "%i"
                },
                .path.a span .crumbs; span itemprop=B>"[a-z]*"; * c@[0] | "%Di\n" / sed "/^$/d" trim "\n",
                .tags.a("|") ul .tagList; a .tag | "%i|" / sed "s/<[^>]*>[^<]*<\/[^>]*>//g; s/|$//"
            """
                )
            )
            dict_add(ret, t)

            rq = reliq(rq.get_data().translate(str.maketrans("", "", "\n\t\r\a")))
            expr = reliq.expr(
                r"""
                h3 .userText; {
                    .user_title em class=b>userTitle | "%i",
                    .user_banner em class=b>userBanner; * c@[0] | "%i" / sed "/^[ \t]*$/d",
                },
                .user_extra dl; {
                    .name dt; * c@[0] | "%Di" / trim,
                    .value dd; * c@[0] | "%i"
                } |
            """
            )
            posts = []

            while True:
                for i in rq.filter(
                    r"ol ( #messageList )( .messageList ); li #E>post-[0-9]* data-author l@[1]"
                ).self():
                    post = {}
                    messageUB = i.filter(r"div class=b>messageUserBlock")

                    avatar, user_id = self.get_avatar_and_userid(ref, messageUB)
                    post["avatar"] = avatar
                    post["user_id"] = user_id

                    t = json.loads(messageUB.search(expr))
                    dict_add(post, t)

                    t = json.loads(
                        i.search(
                            r"""
                        .date div .messageMeta; span .item; [0] * .DateTime; {
                            * l@[0] title | "%(title)v",
                            * l@[0] -title | "%i"
                        },
                        .user li data-author l@[0] | "%(data-author)Dv" / trim,
                        .id.u li data-author l@[0] #B>post-[0-9]* | "%(id)v" / sed "s/^post-//",
                        .text div .messageContent; article | "%i"
                    """
                        )
                    )
                    dict_add(post, t)

                    posts.append(post)

                page += 1
                if (
                    settings["thread_pages_max"] != 0
                    and page >= settings["thread_pages_max"]
                ):
                    break
                nexturl = self.get_next(ref, rq)
                if nexturl is None:
                    break
                rq, ref = self.session.get_html(nexturl, settings, state, True)

            ret["posts"] = posts
            return ret

    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = identify_xenforo1

        self.thread = self.Thread(self.session)
        self.thread.get_next = self.get_next

        self.board_forums_expr = reliq.expr(
            r'h3 .nodeTitle; a -href=a>"#" href | "%(href)v\n"'
        )
        self.forum_forums_expr = self.board_forums_expr
        self.forum_threads_expr = reliq.expr(
            r'li id; div .titleText; h3 .title; a -.prefixLink href | "%(href)v\n"'
        )
        self.guesslist = guesslist

        self.findroot_expr = reliq.expr(
            r"""
            {
                div #navigation; {
                    * .navTab .forums,
                    * .navTab .home
                }; [0] a href | "%(href)v\n",
                fieldset .breadcrumb; a .crumb href; {
                    [0] a C@'span itemprop="title" m@i>"forums"' | '%(href)v\n',
                    [0] a | "%(href)v\n"
                }
            } / line [0] tr "\n"
            """
        )
        self.findroot_board = True
        self.findroot_board_expr = None

    def get_next_page(self, rq):
        url = rq.search(
            r'nav; [0] a class href=Be>"/page-[0-9]*" m@B>"^[^0-9]*&gt;" | "%(href)v"'
        )
        if not re.search(r"/page-[0-9]+$", url):
            return ""
        return url

    def get_tag(self, url, rq=None, state=None, depth=0, **kwargs):
        return self.get_forum(url, rq, state, depth, **kwargs)

    def process_board_r(self, url, ref, rq, settings, state):
        return self.process_forum_r(url, ref, rq, settings, state)

    def process_forum_r(self, url, ref, rq, settings, state):
        t = json.loads(
            rq.search(
                r"""
                .categories { ol #forums; li child@ || ol .nodeList }; {
                    [0] div .nodeInfo child@; div .categoryText; {
                        * .nodeTitle; a; {
                            .name * self@ | "%Di" / trim,
                            .link * self@ | "%(href)v"
                        },
                        .description * .nodeDescription | "%i",
                    },
                    .forums {
                        ol .nodeList l@[:2]; li child@,
                        * self@ -C@"[0] ol .nodeList l@[1:2]"
                    }; div .nodeInfo [0] child@; {
                        .state span .nodeIcon title child@ | "%(title)v",
                        .icon span .nodeIcons; [0] img | "%(src)v",
                        div .nodeText child@; {
                            * .nodeTitle; [0] a; {
                                .name * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            },
                            .description * #b>nodeDescription child@ | "%i",
                            div .nodeStats; dd; {
                                .topics.u [0] * self@ | "%i" tr ",. ",
                                .posts.u [1] * self@ | "%i" tr ",. "
                            },
                        },
                        .feed div .nodeControls; a .feedIcon | "%(href)v",
                        .childboards li .node .level-n; a; {
                            .name * self@ | "%Di" / trim,
                            .link * self@ | "%(href)v"
                        } | ,
                        .lastpost div .nodeLastPost; {
                            * .lastThreadTitle; [0] a; {
                                .title * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            },
                            * .lastThreadMeta; {
                                * .lastThreadUser; [0] a; {
                                    .user * self@ | "%Di" trim,
                                    .user_link * self@ | "%(href)v"
                                },
                                .date * .DateTime | "%(title)v\a%(data-time)v" sed "s/^\a//g; s/\a.*//;"
                            }
                        }
                    } |
                } | ,
                .threads ol .discussionListItems; li id=b>thread-; {
                    .avatar div .posterAvatar; [0] img | "%(src)v",
                    div .main child@; div .titleText child@; {
                        .icons.a div .iconKey child@; span class child@ | "%(class)v\n" / tr " " "\n",
                        * .title child@; {
                            .label [0] a .prefixLink; * c@[0] | "%i",
                            [0] a -.prefixLink; {
                                .title * self@ | "%Di" / trim,
                                .link * self@ | "%(href)v"
                            }
                        },
                        [0] a .username; {
                            .user * self@; [0] * c@[0] | "%Di" trim,
                            .user_link * self@ | "%(href)v"
                        },
                        .date * .DateTime | "%(title)v\a%(data-time)v\a%i" sed "s/^\a//g; s/\a.*//;",
                        .lastpage.u * .itemPageNav; [-] a href | "%i"
                    },
                    div .stats child@; {
                        .replies.u [0] dd | "%i" tr "., ",
                        .views.u [1] dd | "%i" tr "., "
                    },
                    .lastpost div .lastPost child@; {
                        [0] a .username; {
                            .user * self@; * c@[0] | "%Di" / trim,
                            .user_link * self@ | "%(href)v"
                        },
                        .date * .DateTime | "%(title)v\a%(data-time)v\a%i" sed "s/^\a//g; s/\a.*//;"
                    }
                } |
                """
            )
        )

        categories = t["categories"]

        for i in categories:
            i["link"] = url_merge(ref, i["link"])

            for j in i["forums"]:
                for g in j["childboards"]:
                    g["link"] = url_merge(ref, g["link"])

                j["link"] = url_merge(ref, j["link"])

                lastpost = j["lastpost"]
                lastpost["link"] = url_merge(ref, lastpost["link"])
                lastpost["user_link"] = url_merge(ref, lastpost["user_link"])

                j["icon"] = url_merge(ref, j["icon"])

        threads = t["threads"]

        for i in threads:
            i["link"] = url_merge(ref, i["link"])
            i["user_link"] = url_merge(ref, i["user_link"])

            lastpost = i["lastpost"]
            lastpost["user_link"] = url_merge(ref, lastpost["user_link"])

        return {
            "format_version": "xenforo-1-forum",
            "url": url,
            "categories": categories,
            "threads": threads,
        }


class xenforo(ForumExtractorIdentify):
    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.v1 = xenforo1(self.session, **kwargs)
        self.v2 = xenforo2(self.session, **kwargs)

        self.extractors = [self.v1, self.v2]



# by Dominik Stanisław Suchora <suchora.dominik7@gmail.com>
# License: GNU GPLv3

import os
import json
import re
from concurrent.futures import ThreadPoolExecutor
from reliq import reliq

from ..utils import strtosha256, get_settings, url_valid, url_merge
from ..net import Session
from ..enums import Outputs
from ..exceptions import *

common_exceptions = (
    AttributeError,
    ValueError,
    IndexError,
    KeyError,
    RequestError,
    AlreadyVisitedError,
)


def write_json(path, data, settings):
    if settings["compress_func"] is None:
        with open(path, "w") as file:
            file.write(json.dumps(data))
    else:
        with open(path, "wb") as file:
            out = settings["compress_func"](json.dumps(data).encode())
            file.write(out)


def handle_error(self, exception, url, settings, for_pedantic=False):
    if isinstance(exception, AlreadyVisitedError):
        return None

    undisturbed = settings["undisturbed"]
    pedantic = False
    if for_pedantic and settings["pedantic"]:
        pedantic = True

    failed = settings["failed"]
    msg = "{} {}".format(url, exception.args[0])

    if failed:
        if isinstance(failed, list):
            failed.append(msg)
        else:
            print(msg, file=failed)

    if not undisturbed and (failed is None or pedantic):
        if isinstance(exception, RequestError):
            raise RequestError(msg)
        raise exception
    return


def get_first_html(
    extractor, url, settings, state, rq=None, ref=None, return_cookies=False
):
    if rq is None:
        return extractor.session.get_html(
            url, settings, state, extractor.trim, return_cookies
        )

    if not isinstance(rq, reliq):
        rq = reliq(rq)

    if ref is None:
        ref = extractor.session.base(rq, url)
    return (rq, ref)


def state_add_url(typekey, url, state, settings):
    if settings["output"] | Outputs.save_urls:
        state["urls"][typekey].append(url)


def item_file_check_r(url, path_format, i_id, output, force):
    path = None

    if path_format and Outputs.write_by_id in output:
        path = path_format.format(i_id)
    elif Outputs.write_by_hash in output:
        path = strtosha256(url)

    if path and not force and os.path.exists(path) and os.path.getsize(path) != 0:
        return
    return path


def item_file_check(url, path_format, i_id, settings):
    return item_file_check_r(
        url, path_format, i_id, settings["output"], settings["force"]
    )


def get_baseurl(url):
    r = url_valid(url, base=True)
    if r is None:
        return ""

    base, rest = r
    p = base.find("//")
    if p == -1:
        return ""
    return base[p + 2 :]


class ItemExtractor:
    def __init__(self, session):
        self.path_format = "{}"
        self.match = [(None, 0)]
        self.trim = False

        self.session = session
        self.common_exceptions = common_exceptions

    def state_add_url(self, keytype, url, state, settings):
        return state_add_url(keytype, url, state, settings)

    def get_url(self, url):
        return url

    def handle_error(self, exception, url, settings, for_pedantic=False):
        return handle_error(self, exception, url, settings, for_pedantic)

    def get_first_html(
        self, url, settings, state, rq=None, ref=None, return_cookies=False
    ):
        return get_first_html(self, url, settings, state, rq, ref, return_cookies)

    def get_improper_url(self, url, rq, settings, state):
        print('improper url - "{}"'.format(url), file=settings["logger"])
        return [None, 0]

    def url_get_id(self, url):
        for i in self.match:
            r = url_valid(url, regex=i[0], base=True, matchwhole=True)
            if r is not None:
                return r[1][i[1]]

    def get(self, typekey, url, settings, state, rq=None):
        outtype = settings["output"]
        if Outputs.only_urls_forums in outtype:
            return

        if Outputs.only_urls_threads in outtype:
            self.state_add_url(typekey, url, state, settings)
            return state

        i_id = self.url_get_id(url)

        if i_id is None:
            rq, i_id = self.get_improper_url(url, rq, settings, state)
            if not rq:
                return

        path = item_file_check(url, self.path_format, i_id, settings)

        if not path and Outputs.data not in outtype:
            return state

        url = self.get_url(url)
        rq, ref = self.get_first_html(url, settings, state, rq)

        contents = self.get_contents(rq, settings, state, url, ref, i_id)
        if contents is None:
            return

        if Outputs.data in outtype:
            self.state_add_url(typekey, url, state, settings)
            state["data"][typekey].append(contents)

        if path:
            write_json(path, contents, settings)
            self.state_add_url(typekey, url, state, settings)
            state["files"][typekey].append(path)

        return state

    def get_contents(self, rq, settings, state, url, ref, i_id):
        pass


class ForumExtractor:
    def __init__(self, session=None, **kwargs):
        self.thread = None
        self.user = None

        self.forum_threads_expr = None
        self.forum_forums_expr = None
        self.tag_threads_expr = None
        self.board_forums_expr = None
        self.guesslist = []
        self.domains = []
        self.domain_guess_mandatory = False

        self.trim = False

        self.identify_func = None

        self.common_exceptions = common_exceptions

        self.settings = {
            "thread_pages_max": 0,
            "pages_max": 0,
            "pages_max_depth": 0,
            "pages_threads_max": 0,
            "pages_forums_max": 0,
            "output": Outputs.write_by_id | Outputs.urls | Outputs.threads,
            "max_workers": 1,
            "undisturbed": False,
            "pedantic": False,
            # requests settings
            "force": False,
            "verify": True,
            "allow_redirects": False,
            "timeout": 120,
            "proxies": {},
            "headers": {},
            "cookies": {},
            # net settings
            "user-agent": None,
            "wait": 0,
            "wait_random": 0,
            "logger": None,
            "failed": None,
            "retries": 3,
            "retry_wait": 60,
            "force": False,
            "compress_func": None,
        }
        self.settings = self.get_settings(kwargs)

        if session:
            self.session = session
        else:
            self.session = Session(**self.settings)

        self.findroot_expr = None
        self.findroot_board = False
        self.findroot_board_expr = None

    def get_settings(self, settings):
        ret = get_settings(self.settings, **settings)

        if (
            Outputs.only_urls_threads in ret["output"]
            or Outputs.only_urls_forums in ret["output"]
        ):
            ret["force"] = True

        if ret["user-agent"] is not None:
            ret["headers"].update({"User-Agent": ret["user-agent"]})

        return ret

    @staticmethod
    def create_state(state):
        if state:
            return state
        return {
            "files": {
                "boards": [],
                "tags": [],
                "forums": [],
                "threads": [],
                "users": [],
            },
            "urls": {
                "threads": [],
                "users": [],
                "reactions": [],
                "forums": [],
                "tags": [],
                "boards": [],
            },
            "data": {
                "boards": [],
                "tags": [],
                "forums": [],
                "threads": [],
                "users": [],
            },
            "visited": set(),
            "scraper": None,
            "scraper-method": None,
        }

    def handle_error(self, exception, url, settings, for_pedantic=False):
        return handle_error(self, exception, url, settings, for_pedantic)

    def get_first_html(
        self, url, settings, state, rq=None, ref=None, return_cookies=False
    ):
        return get_first_html(self, url, settings, state, rq, ref, return_cookies)

    def _get_item(self, typekey, url, obj, rq=None, state=None, depth=0, **kwargs):
        settings = self.get_settings(kwargs)
        state = self.create_state(state)
        try:
            return obj.get(typekey, url, settings, state, rq)
        except self.common_exceptions as ex:
            return self.handle_error(ex, url, settings)

    def get_thread(self, url, rq=None, state=None, depth=0, **kwargs):
        return self._get_item("threads", url, self.thread, rq, state, depth, **kwargs)

    def get_user(self, url, rq=None, state=None, depth=0, **kwargs):
        return self._get_item("users", url, self.user, rq, state, depth, **kwargs)

    get_next_page = None  # empty functions
    get_forum_next_page = None
    get_tag_next_page = None

    def get_next(self, ref, rq):
        return url_merge(ref, self.get_next_page(rq))

    def get_forum_next(self, ref, rq):
        if self.get_forum_next_page is None:
            return self.get_next(ref, rq)
        return url_merge(ref, self.get_forum_next_page(rq))

    def get_tag_next(self, ref, rq):
        if self.get_tag_next_page is None:
            return self.get_next(ref, rq)
        return url_merge(ref, self.get_tag_next_page(rq))

    def go_through_page_thread(self, url, settings, state, depth):
        try:
            return self.get_thread(url, None, state, depth + 1, **settings)
        except self.common_exceptions as ex:
            return self.handle_error(ex, url, settings)

    def go_through_page_threads(self, ref, rq, settings, state, expr, depth):
        if (
            Outputs.threads not in settings["output"]
            and Outputs.only_urls_threads not in settings["output"]
        ):
            return
        urls = rq.search(expr).split("\n")[:-1]
        urls_len = len(urls)
        if (
            settings["pages_threads_max"] > 0
            and urls_len >= settings["pages_threads_max"]
        ):
            urls_len = settings["pages_threads_max"]
            urls = urls[:urls_len]

        urls_len = len(urls)
        urls = list(map(lambda x: url_merge(ref, x), urls))

        if settings["max_workers"] > 1 and urls_len > 0:
            with ThreadPoolExecutor(max_workers=settings["max_workers"]) as executor:
                list(
                    executor.map(
                        lambda x: self.go_through_page_thread(
                            x, settings, state, depth
                        ),
                        urls,
                    )
                )
        else:
            for url in urls:
                self.go_through_page_thread(url, settings, state, depth)

    def go_through_page_forums(self, ref, rq, settings, state, expr, depth):
        max_forums = settings["pages_forums_max"]

        for num, url in enumerate(rq.search(expr).split("\n")[:-1]):
            if max_forums > 0 and num >= max_forums:
                break

            url = url_merge(ref, url)

            if (
                settings["pages_max_depth"] == 0
                or settings["pages_max_depth"] > depth + 1
            ):
                try:
                    self.get_forum(url, None, state, depth + 1, **settings)
                except (RequestError, AlreadyVisitedError):
                    continue

    def go_through_pages(
        self,
        typekey,
        url,
        threads_expr,
        forums_expr,
        func_next,
        depth,
        rq=None,
        state=None,
        process_func=None,
        **kwargs,
    ):
        settings = self.get_settings(kwargs)
        state = self.create_state(state)
        try:
            rq, ref = self.get_first_html(url, settings, state, rq)
        except self.common_exceptions as ex:
            return self.handle_error(ex, url, settings)

        page = 0
        state_add_url(typekey, url, state, settings)

        if process_func:
            process_func(url, ref, rq, settings, state)

        if forums_expr:
            self.go_through_page_forums(ref, rq, settings, state, forums_expr, depth)

        if (
            threads_expr
            and (
                Outputs.threads in settings["output"]
                or Outputs.only_urls_threads in settings["output"]
                or Outputs.forums in settings["output"]
                or Outputs.tags
            )
            and Outputs.only_urls_forums not in settings["output"]
        ):
            while True:
                self.go_through_page_threads(
                    ref, rq, settings, state, threads_expr, depth
                )

                if not func_next:
                    break

                page += 1
                if settings["pages_max"] > 0 and page >= settings["pages_max"]:
                    break
                nexturl = func_next(ref, rq)
                if nexturl is None:
                    break
                try:
                    rq, ref = self.get_first_html(nexturl, settings, state)
                except self.common_exceptions as ex:
                    self.handle_error(ex, nexturl, settings)
                    break

                if process_func:
                    process_func(nexturl, ref, rq, settings, state)

        return state

    def process_forum_r(self, url, ref, rq, settings, state):
        pass

    def process_tag_r(self, url, ref, rq, settings, state):
        pass

    def process_board_r(self, url, ref, rq, settings, state):
        pass

    def process_page(
        self, url, ref, rq, settings, state, prefix, typekey, process_func, otype
    ):
        output = settings["output"]
        if otype not in output or (
            output & Outputs.writers == 0 and Outputs.data not in output
        ):
            return

        path = None

        if output & Outputs.writers != 0:
            path = item_file_check_r(
                url, "", "0", Outputs.write_by_hash, settings["force"]
            )
            if path is None:
                return
            path = "{}-{}".format(prefix, path)

        data = process_func(url, ref, rq, settings, state)

        if Outputs.data in output:
            state["data"][typekey].append(data)
        if path:
            write_json(path, data, settings)
            state["files"][typekey].append(path)

    def process_forum(self, url, ref, rq, settings, state):
        return self.process_page(
            url,
            ref,
            rq,
            settings,
            state,
            "f",
            "forums",
            self.process_forum_r,
            Outputs.forums,
        )

    def process_tag(self, url, ref, rq, settings, state):
        return self.process_page(
            url,
            ref,
            rq,
            settings,
            state,
            "t",
            "tags",
            self.process_tag_r,
            Outputs.tags,
        )

    def process_board(self, url, ref, rq, settings, state):
        return self.process_page(
            url,
            ref,
            rq,
            settings,
            state,
            "b",
            "boards",
            self.process_board_r,
            Outputs.boards,
        )

    def get_forum(self, url, rq=None, state=None, depth=0, process_func=None, **kwargs):
        if process_func is None:
            process_func = self.process_forum
        return self.go_through_pages(
            "forums",
            url,
            self.forum_threads_expr,
            self.forum_forums_expr,
            self.get_forum_next,
            depth,
            rq,
            state,
            process_func,
            **kwargs,
        )

    def get_tag(self, url, rq=None, state=None, depth=0, **kwargs):
        if not self.tag_threads_expr:
            return None
        return self.go_through_pages(
            "tags",
            url,
            self.tag_threads_expr,
            None,
            self.get_tag_next,
            depth,
            rq,
            state,
            self.process_tag,
            **kwargs,
        )

    def get_board(self, url, rq=None, state=None, depth=0, **kwargs):
        if not self.board_forums_expr:
            return self.get_forum(url, rq, state, depth, self.process_board, **kwargs)
        return self.go_through_pages(
            "boards",
            url,
            None,
            self.board_forums_expr,
            None,
            depth,
            rq,
            state,
            self.process_board,
            **kwargs,
        )

    def guess_in_domain(self, url):
        base = get_baseurl(url)
        if base == "":
            return

        if base in self.domains:
            return self

    def guess_search(self, url, found_domain, **kwargs):
        if found_domain is False:
            d = self.guess_in_domain(url)
            if d is not None:
                c = d.guess(url, found_domain=True)
                return c

            if self.domain_guess_mandatory is True:
                print(
                    "url {} was not found in domain list, refusing to guess".format(
                        url
                    ),
                    file=self.get_settings(kwargs)["logger"],
                )
                return

        func = None
        for i in self.guesslist:
            func = getattr(self, i["func"])
            exprs = i["exprs"]

            if exprs:
                for expr in exprs:
                    if re.search(expr, url):
                        return func
            else:
                return func
        return

    def guess(self, url, rq=None, state=None, found_domain=False, **kwargs):
        rest = url_valid(url)
        if rest is None:
            return

        state = self.create_state(state)
        func = self.guess_search(url, found_domain, **kwargs)

        if found_domain:
            return func

        if func:
            state["scraper-method"] = func
            return func(url, rq=rq, state=state, **kwargs)
        return

    def findroot_r(self, url, ref, rq, teststate, settings):
        rest = url_valid(url, base=True)
        if rest is None:
            return

        if self.findroot_board:
            settings["output"] = Outputs.only_urls_forums
            settings["pages_max_depth"] = 1
            self.get_board(
                url,
                rq,
                teststate,
                **settings,
            )
            if len(teststate["urls"]["forums"]) > 0 and (
                not self.findroot_board_expr
                or (
                    len(rest) == 0
                    or rest == "/"
                    or re.search(
                        self.findboard_board_expr,
                        rest,
                    )
                )
            ):
                return url

        if self.findroot_expr is None:
            return

        newurl = rq.search(self.findroot_expr)
        return url_merge(ref, newurl)

    def findroot(self, url, rq=None, state=None, **kwargs):
        settings = self.get_settings(kwargs)
        teststate = self.create_state(None)

        try:
            rq, ref = self.get_first_html(url, settings, teststate, rq)
        except self.common_exceptions as ex:
            return self.handle_error(ex, url, settings)

        url = self.findroot_r(url, ref, rq, teststate, settings)
        if url is None:
            return

        url = re.sub(r"[?&](sid|PHPSESSID)=[a-zA-Z0-9]+", r"", url)
        index1 = url.find("?")
        index2 = url.find("&")
        if index1 == -1 and index2 != -1:
            url = url[:index2] + "?" + url[(index2 + 1) :]
        url = re.sub(r"&", r"?", url, count=1)
        if url[-1] == "?":
            url = url[:-1]
        return url

    def identify_page(self, url, rq, cookies):
        if self.identify_func is None:
            return

        d = self.guess_in_domain(url)
        if d is not None:
            return d

        if self.domain_guess_mandatory is True:
            return

        r = self.identify_func(url, rq, cookies)
        if r is True:
            return self
        if r is False:
            return
        return r


class ForumExtractorIdentify(ForumExtractor):
    def __init__(self, session=None, **kwargs):
        super().__init__(session, **kwargs)

        self.identify_func = self.listIdentify

        self.extractors = []

    def guess_in_domain(self, url):
        base = get_baseurl(url)
        if base == "":
            return

        for i in self.extractors:
            r = i.guess_in_domain(url)

            if r is not None:
                return r

    def listIdentify(self, url, rq, cookies):
        for i in self.extractors:
            if i.identify_func is None:
                continue

            r = i.identify_page(url, rq, cookies)
            if r is not None:
                return r

    def get_unknown(self, func_name, url, rq=None, state=None, **kwargs):
        settings = self.get_settings(kwargs)

        if Outputs.write_by_hash in settings["output"] and (
            func_name == "get_thread" or func_name == "get_user"
        ):
            if not item_file_check(url, None, None, settings):
                return

        state = self.create_state(state)

        try:
            r = self.get_first_html(url, settings, state, rq, return_cookies=True)
            cookies = {}
            if rq is None:
                cookies = r[2]
            rq = r[0]
            ref = r[1]
        except self.common_exceptions as ex:
            return self.handle_error(ex, url, settings)

        forum = self.identify_page(url, rq, cookies)
        if forum is None:
            return
        state["scraper"] = forum

        if func_name is None:
            return state

        func = getattr(forum, func_name)
        state["scraper-method"] = func
        return func(url, rq, state, **settings)

    def guess(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("guess", url, rq, state, **kwargs)

    def get_thread(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("get_thread", url, rq, state, **kwargs)

    def get_user(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("get_user", url, rq, state, **kwargs)

    def get_forum(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("get_forum", url, rq, state, **kwargs)

    def get_tag(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("get_tag", url, rq, state, **kwargs)

    def get_board(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("get_board", url, rq, state, **kwargs)

    def identify(self, url, rq=None, state=None, **kwargs):
        r = self.get_unknown(None, url, rq, None, **kwargs)
        if r is None:
            return
        return r["scraper"]

    def findroot(self, url, rq=None, state=None, **kwargs):
        return self.get_unknown("findroot", url, rq, None, **kwargs)
